/////////////////////////////////////////////////////////////////////////////
// Name:        MyDirs.cpp
// Purpose:    Dir-view
// Part of:       4Pane
// Author:       David Hart
// Copyright:   (c) 2007 David Hart
// Licence:     GPL v3
/////////////////////////////////////////////////////////////////////////////
// For compilers that support precompilation, includes "wx/wx.h".  
#include "wx/wxprec.h" 

#ifndef WX_PRECOMP
    #include "wx/log.h"
    #include "wx/app.h"
    #include "wx/frame.h" 

    #include "wx/scrolwin.h"
    #include "wx/menu.h"
	#include <wx/dirctrl.h>
	#include "wx/dcmemory.h"
	#include "wx/dragimag.h"
	#include "wx/config.h"
#endif
 
#include "wx/splitter.h"
#include <wx/file.h>


#include "MyTreeCtrl.h"
#include "MyFiles.h"
#include "Filetypes.h"
#include "Accelerators.h"
#include "Redo.h"
#include "MyDirs.h"
#include "MyFrame.h"
#include "Configure.h"
#include "Misc.h"

#include "bitmaps/include/fulltree.xpm"
#include "bitmaps/include/up.xpm"			


class MakeLinkDlg	:	public wxDialog
{
public:
~MakeLinkDlg();
void init( bool multiple, bool dirsISQ );

wxStaticText* frompath;
wxStaticText* topath;
wxTextCtrl* LinkExt;
wxTextCtrl* LinkFilename;
wxCheckBox* MakeRelative;
wxCheckBox* ApplyToAll;
bool CanApply2All;
bool samedir;
long lastradiostation;

protected:
void OnLinkRadioButton(wxCommandEvent& event);		 // If choose-name radio clicked
void OnRelativeCheckUI(wxUpdateUIEvent& event);
void OnSkip(wxCommandEvent& WXUNUSED(event)){ EndModal( XRCID("Skip") ); }

wxRadioBox* Linktype;
wxString lastext;
DECLARE_EVENT_TABLE()
};


void MakeLinkDlg::init( bool multiple, bool dirsISQ )
{
CanApply2All = multiple; samedir = dirsISQ;
frompath = (wxStaticText*)FindWindow( wxT("link_frompath") );
topath = (wxStaticText*)FindWindow( wxT("link_path") );
LinkExt = (wxTextCtrl*)FindWindow( wxT("link_ext") );
LinkFilename = (wxTextCtrl*)FindWindow( wxT("link_filename") );
Linktype = (wxRadioBox*)FindWindow( wxT("link_radiobox") );
MakeRelative = (wxCheckBox*)FindWindow( wxT("MakeRelCheck") );
ApplyToAll = (wxCheckBox*)FindWindow( wxT("ApplyToAllCheck") ); ApplyToAll->Enable(CanApply2All);

wxConfigBase::Get()->Read( wxT("History/MakeLinkDlg/Ext"), &lastext, wxT("") ); LinkExt->SetValue( lastext );
lastradiostation = wxConfigBase::Get()->Read( wxT("History/MakeLinkDlg/Filename"), 0l );
if ( samedir && lastradiostation==0 ) lastradiostation = 1;						// If source & dest dirs are the same, we can't retain the same name

bool rel; wxConfigBase::Get()->Read( wxT("History/MakeLinkDlg/MakeRelative"), &rel, false ); MakeRelative->SetValue( rel );

wxButton* skip = (wxButton*)FindWindow( wxT("Skip") );
if ( skip ) skip->Show( multiple );																// Show the Skip button only if there are multiple targets

int ID;
switch( lastradiostation )		// How did we last make a name for the link:  reuse the original, add an ext, or bespoke?
	{	case 0: ((wxRadioButton*)FindWindow( wxT("NameSame") ))->SetValue(true); ID=XRCID("NameSame"); break;
		case 1: ((wxRadioButton*)FindWindow( wxT("NameExt") ))->SetValue(true); ID=XRCID("NameExt");  break;
		default: ((wxRadioButton*)FindWindow( wxT("NameName") ))->SetValue(true); ID=XRCID("NameName"); 
	}

wxCommandEvent event; event.SetId( ID ); OnLinkRadioButton(event);	// Fake an event to get the UpdateUI right
}

MakeLinkDlg::~MakeLinkDlg()
{
wxConfigBase::Get()->Write( wxT("History/MakeLinkDlg/Filename"), lastradiostation );
if ( ! LinkExt->GetValue().IsEmpty() )  wxConfigBase::Get()->Write( wxT("History/MakeLinkDlg/Ext"), LinkExt->GetValue() );
if ( MakeRelative->IsEnabled() ) wxConfigBase::Get()->Write( wxT("History/MakeLinkDlg/MakeRelative"), MakeRelative->IsChecked() );
}

void MakeLinkDlg::OnLinkRadioButton(wxCommandEvent& event)		// UI if name-type radiobuttons clicked in Make Link dialog
{
if ( event.GetId() != XRCID("NameName") )														// Do buttons 1 & 2 first
	{	bool Ext = (event.GetId() == XRCID("NameExt"));
		if ( samedir && ! Ext )
			{	BriefMessageBox msg( _("You can't link within a directory unless you alter the link's name."), 3, _("Not possible") );
				((wxRadioButton*)FindWindow( wxT("NameName") ))->SetValue( true ); return;
			}
		LinkFilename->Disable(); LinkExt->Enable( Ext ); lastradiostation = Ext;
		ApplyToAll->Enable(CanApply2All);  return;												// Can ApplyToAll, providing there are >1 items
	}

LinkFilename->Enable(); LinkExt->Disable(); lastradiostation = 2;						// OK, button 3 was selected
if ( LinkFilename->GetValue().IsEmpty() ) LinkFilename->SetValue( frompath->GetLabel().AfterLast(wxFILE_SEP_PATH) ); // Use the original name as template for the new one
ApplyToAll->Disable(); 																						// Can't ApplyToAll, even if there are >1 items
}

void MakeLinkDlg::OnRelativeCheckUI(wxUpdateUIEvent& event)
{
if ( Linktype ) MakeRelative->Enable( Linktype->GetSelection() );	// This enables the button only if it's a symlink that is going to be made
}

BEGIN_EVENT_TABLE(MakeLinkDlg,wxDialog)
	EVT_RADIOBUTTON(wxID_ANY, MakeLinkDlg::OnLinkRadioButton)
	EVT_BUTTON(XRCID("Skip"), MakeLinkDlg::OnSkip)
	EVT_UPDATE_UI(XRCID("MakeRelCheck"), MakeLinkDlg::OnRelativeCheckUI)
END_EVENT_TABLE()

//-----------------------------------------------------------------------------------------------------------------------

IMPLEMENT_DYNAMIC_CLASS(MyDropdownButton, wxBitmapButton)

void MyDropdownButton::OnButtonClick(wxCommandEvent& event)
{
int width, height;
GetClientSize(&width, &height);														// How big are we?
wxPoint clientpt, screenpt = wxGetMousePosition();					// Get screen position
clientpt = ScreenToClient( screenpt );											// & from that get the position within the button
screenpt = ((wxToolBar*)GetParent())->ScreenToClient( screenpt );											// & from that get the position within the button

left = screenpt.x - clientpt.x;															// The left of the button is screen_pos - position_within_button
right = left + width;																			// The right is left + width

left -= partnerwidth;																			// For left dirviews, we actually want the left edge of the PREVIOUS button

event.Skip();
}

BEGIN_EVENT_TABLE(MyDropdownButton, wxBitmapButton)
	EVT_BUTTON(-1, MyDropdownButton::OnButtonClick)
END_EVENT_TABLE()

//-----------------------------------------------------------------------------------------------------------------------

DirGenericDirCtrl::DirGenericDirCtrl( wxWindow* parent, const wxWindowID id, const wxString& START_DIR , const wxPoint& pos, const wxSize& size,
																		 long style, bool full_tree, const wxString& name )
			:  MyGenericDirCtrl( parent, (MyGenericDirCtrl*)this, id, START_DIR ,  pos,  size ,  style | wxDIRCTRL_DIR_ONLY, wxT(""), 0, name, ISLEFT, full_tree )
{
if ( !startdir.IsEmpty() )																		// Store initial startdir in array, so that if we use back/forward buttons, it's there to return to
	{	if ( startdir.Last() == wxFILE_SEP_PATH ) PreviousDirs.Add( startdir.BeforeLast(wxFILE_SEP_PATH) );
			else PreviousDirs.Add( startdir );											// We want it without the terminal '/' (which it probably has), lest future comparisons fail
	}
count=0;																								// Set array count to 0 to point to the above entry
toolBar = NULL;


CreateAcceleratorTable();
}

/*
enum
{
    wxACCEL_NORMAL  = 0x0000,   // no modifiers
    wxACCEL_ALT     = 0x0001,   // hold Alt key down
    wxACCEL_CTRL    = 0x0002,   // hold Ctrl key down
    wxACCEL_SHIFT   = 0x0004    // hold Shift key down
};
*/

DirGenericDirCtrl::~DirGenericDirCtrl()
{ 
if (toolBar != NULL)  toolBar->Destroy();
}

void DirGenericDirCtrl::CreateAcceleratorTable()
{
int AccelEntries[] = { SHCUT_CUT, SHCUT_COPY, SHCUT_PASTE, SHCUT_HARDLINK, SHCUT_SOFTLINK, SHCUT_TRASH,
										SHCUT_DELETE, SHCUT_RENAME, SHCUT_NEW, SHCUT_UNDO, SHCUT_REDO, IDM_TOOLBAR_fulltree,
										SHCUT_REFRESH, SHCUT_FILTER, SHCUT_TOGGLEHIDDEN, SHCUT_PROPERTIES };
const size_t shortcutNo = sizeof(AccelEntries)/sizeof(int);
MyFrame::mainframe->AccelList->CreateAcceleratorTable( this, AccelEntries,  shortcutNo );
}

void DirGenericDirCtrl::RecreateAcceleratorTable()
{
				// It would be nice to be able to empty the old accelerator table before replacing it, but trying caused segfaulting!
CreateAcceleratorTable();																					// Make a new one, with up-to-date data
((MyTreeCtrl*)GetTreeCtrl())->RecreateAcceleratorTable();								// Tell the tree to do likewise
}
	 
void DirGenericDirCtrl::CreateToolbar()
{
long style =  wxTB_HORIZONTAL;

DirSplitterWindow* DSW;													// Get ptr to the parent of the toolbar
if ( isright )
			DSW = ParentTab->m_splitterRightBottom;	
 else	DSW = ParentTab->m_splitterLeftTop;
 
if ( toolBar != NULL ) toolBar->Destroy();							// In case we're recreating

toolBar = new wxToolBar(DSW, ID_TOOLBAR,wxDefaultPosition, wxDefaultSize, style);

wxBitmap toolBarBitmaps[2];


toolBarBitmaps[0] = wxBitmap(fulltree_xpm);						// First the "standard" buttons
toolBarBitmaps[1] = wxBitmap(up_xpm);
																							// Load the navigation buttons from XRC
wxBitmapButton* backbtn = (wxBitmapButton*)wxXmlResource::Get()->LoadObject(toolBar, wxT("IDM_TOOLBAR_back"), wxT("wxBitmapButton"));
wxBitmapButton* forwardbtn = (wxBitmapButton*)wxXmlResource::Get()->LoadObject(toolBar, wxT("IDM_TOOLBAR_forward"), wxT("wxBitmapButton"));
MenuOpenButton1 = (MyDropdownButton*)wxXmlResource::Get()->LoadObject(toolBar, wxT("IDM_TOOLBAR_smalldropdown1"), wxT("wxBitmapButton"));
MenuOpenButton2 = (MyDropdownButton*)wxXmlResource::Get()->LoadObject(toolBar, wxT("IDM_TOOLBAR_smalldropdown2"), wxT("wxBitmapButton"));

int width, height; backbtn->GetClientSize(&width, &height);
MenuOpenButton1->partnerwidth = width;							// Tell the dropdown buttons how wide is their neighbour:  needed correctly to position the dropdown menus
MenuOpenButton2->partnerwidth = width;

toolBar->AddTool(IDM_TOOLBAR_fulltree, wxT("Full Tree"), toolBarBitmaps[0], _("Show Full Tree"), wxITEM_CHECK);
toolBar->AddTool(IDM_TOOLBAR_up, wxT("Up"), toolBarBitmaps[1], _("Up to Higher Directory"), wxITEM_NORMAL);
toolBar->AddControl( backbtn );
toolBar->AddControl( MenuOpenButton1 );
toolBar->AddControl( forwardbtn );
toolBar->AddControl( MenuOpenButton2 );
toolBar->AddSeparator();
#ifndef __WXGTK20__
toolBar->AddSeparator();													// 2 separators look prettier in gtk 1.2
#endif
																							// Now load any user-prefs buttons eg Home, BM1
ArrayofDirCtrlTBStructs* array = MyFrame::mainframe->configure->DirCtrlTBinfoarray;	// Readability
size_t count = wxMin( array->GetCount(), IDM_TOOLBAR_Range );	// Check there aren't too many user-defined buttons requested
for ( size_t n=0; n < count; ++n )
	toolBar->AddTool(IDM_TOOLBAR_bmfirst + n, wxEmptyString, array->Item(n)->bitmaplocation, array->Item(n)->tooltip, wxITEM_NORMAL);

toolBar->Realize(); 																// Essential to make the tools display

DSW->DirToolbarSizer->Add( toolBar, 1, wxEXPAND );
DSW->DirToolbarSizer->Layout();										// Needed for early wx versions
}

void DirGenericDirCtrl::OnDirCtrlTBButton( wxCommandEvent& event )		// When a user-defined dirctrlTB button is clicked
{
size_t buttonnumber = event.GetId() - IDM_TOOLBAR_bmfirst;			// Find which button was clicked
if ( buttonnumber >= MyFrame::mainframe->configure->DirCtrlTBinfoarray->GetCount() ) return; 	// Sanity check

wxString path = MyFrame::mainframe->configure->DirCtrlTBinfoarray->Item(buttonnumber)->destination;

if ( ((MyGenericDirCtrl*)this)->arcman->NavigateArchives( path ) == Snafu )  return;		// This copes with both archives and ordinary filepaths

startdir = path; RefreshTree( path );
SetPath( path );
UpdateStatusbarInfo( path ); AddPreviouslyVisitedDir( path );
}

void DirGenericDirCtrl::DoToolbarUI(wxUpdateUIEvent& event)	//  Greys out appropriate buttons 
{
switch( event.GetId() )
	{	case IDM_TOOLBAR_fulltree:	event.Check( fulltree ); 				// Set the display according to the bool
															toolBar->EnableTool( IDM_TOOLBAR_fulltree,  ! arcman->IsArchive() ); return;	// Disable if within an archive
		case IDM_TOOLBAR_up: 	event.Enable( ! ( fulltree || startdir.IsSameAs(wxFILE_SEP_PATH) ) ); return; // Enable Up button if not fulltree && startdir!=root 
	}
#if ! defined( __WXX11__ )	   			// For some reason, this crashes on X11 with an invalid bitmap error message !?
if  ( event.GetId()==XRCID("IDM_TOOLBAR_back") ) { event.Enable( count > 0 ); MenuOpenButton1->Enable(count > 0 ); return; }// Enable dir buttons together with their sidebars
if ( event.GetId()==XRCID("IDM_TOOLBAR_forward") ) { event.Enable( (count+1) < PreviousDirs.GetCount() ); MenuOpenButton2->Enable( (count+1) < PreviousDirs.GetCount() ); }
#endif
}

void DirGenericDirCtrl::OnFullTree( wxCommandEvent& WXUNUSED(event) )	//	When FullTree button clicked
{
if ( arcman->IsArchive() ) return;															// We don't want to allow fulltree-mode if inside an archive

fulltree = ! fulltree;																					// Toggle it

startdir = GetPath();																				// Rebase startdir to current selection
if ( !fulltree )    AddPreviouslyVisitedDir( startdir );									// If we're changing from fulltree to branch mode, store the new rootpath for posterity 
ReCreateTreeFromSelection();																// Rebase the tree, from selection or from root
}

void DirGenericDirCtrl::OnDirUp(wxCommandEvent& WXUNUSED(event))			// From toolbar Up button.  Go up a directory (if not in Fulltree mode)
{
if ( fulltree ) return;																				// It doesn't make sense if we're already displaying the whole tree

wxString path = startdir;

while ( path.Last()==wxFILE_SEP_PATH && path.Len() > 1 ) 				// Unlikely, but remove multiple terminal '/'s
		path.RemoveLast();														

if ( path.IsSameAs(wxFILE_SEP_PATH) )   return;								// Can't get higher than root!

if ( path.Last()==wxFILE_SEP_PATH ) 	path.RemoveLast();				// If there's a terminal '/', remove it first
startdir = path.BeforeLast(wxFILE_SEP_PATH);									// Now truncate path by one segment & put in startdir
startdir += wxFILE_SEP_PATH;																// Add back the terminal '/', partly because otherwise going up from "/home" would give ""!

AddPreviouslyVisitedDir( startdir );															// Store the new startdir for posterity
RefreshTree( startdir );																			// Finally use RefreshTree to redo the tree
}

void DirGenericDirCtrl::OnDirBack(wxCommandEvent& WXUNUSED(event))		// From toolbar Back button.  Go back to a previously-visited directory
{
RetrieveEntry( -1 );																					// Go back one
}

void DirGenericDirCtrl::OnDirForward(wxCommandEvent& WXUNUSED(event))	// From toolbar Forward button.  Go forward to a previously-undone directory
{
RetrieveEntry( 1 );																					// Go forward one
}

void DirGenericDirCtrl::OnDirDropdown(wxCommandEvent& event)	// When one of the Previous/Next Directory dropdown buttons is clicked
{
int x; bool rightside, back;

MyDropdownButton* button =  (MyDropdownButton*)event.GetEventObject();
back = (event.GetId() == XRCID("IDM_TOOLBAR_smalldropdown1"));	// See if the click was on the back or forwards dropdown button

rightside = isright && (ParentTab->splitstatus==vertical);							// Do things differently for right-side dirviews, but only if split vertically!

if ( rightside )
	x = button->right;																					// The edge of the popdown menu will be on the Rt edge of the dropdown button
  else
	x = button->left;																						//   or on the left edge of its left neighbour

wxPoint location(x, 0);
MyPopupMenu(this, location, back, rightside);											// This is a class that does the pop-up menu holding previously visited dirs to select.  If 'back' we're going backward
}

void DirGenericDirCtrl::AddPreviouslyVisitedDir( wxString& newentry )		// Adds a new revisitable dir to the array.  NB count indexes currently-displayed entry
{
if ( newentry.IsEmpty() ) return;											// Some implausible problem occurred

if ( count == MAX_NUMBER_OF_PREVIOUSDIRS )
	{	PreviousDirs.RemoveAt( (size_t)0 ); --count; }			// Oops: too many entries.  Do a FIFO by removing oldest entry

			// Check that there haven't already been some undos.  If so, adding another undo invalidates any redos, so delete them
size_t arrno = PreviousDirs.GetCount();								// Get the real size of array
if ( (count+1) < arrno )														// If it's bigger, there must be redundant entries
	PreviousDirs.RemoveAt( count+1, arrno-(count+1) );		// Delete any entries > count

if ( (arrno=PreviousDirs.GetCount()) > 0 && arrno > count	// Needed to avoid out-of-range segfaults
			&& PreviousDirs[count] == newentry )   return;	// In case we've been navigating in a circle; we don't want the back-button make us march on the spot

PreviousDirs.Add( newentry ); 											// Append new entry
++count;																				// Finally inc the count
}


void DirGenericDirCtrl::RetrieveEntry( int displacement )		// Retrieves a dir from the array.  'count + displacement' determines which. 
{																								// 'displacement' may be + or -, depending on whether we're going backwards or forwards
if ( !PreviousDirs.GetCount() ) return;

if ( displacement > 0 )																	// If +ve, we're going forwards
	{	if ( (count+displacement) > PreviousDirs.GetCount() ) 		// Check we're within bounds
			return;
	}
 else																							// Otherwise we're going backwards
 	{	if ( ((size_t)(-displacement)) > count )	return;	}				// Displacement will be a minus figure, probably -1.  Check it's not TOO minus!

count += displacement;																// Move the index to requested entry. Note the += works even if displacement is negative
wxString entry = PreviousDirs[count];											// Extract the relevant path
if ( entry.IsEmpty() )  return;

if ( (MyGenericDirCtrl*)arcman == NULL ) return;

if ( ((MyGenericDirCtrl*)this)->arcman->NavigateArchives( entry ) != Snafu )	// This copes with both archives and ordinary filepaths
	{	startdir = entry; RefreshTree( entry );	return; }	// Display new startdir

wxMessageBox(_("Hmm.  It seems that directory no longer exists!"));	// If things weren't valid, scream & shout
if ( PreviousDirs.GetCount() > 1 )  								// Make sure there's at least 1 more dir stored
	{	PreviousDirs.RemoveAt( count );						// If so, delete the dud entry
		count -= displacement;										// Return count to where it started
		if ( displacement < 0 ) --count;							//   & dec it if needed to compensate for the deleted entry.  It should now index the current dir again
		return;
	}

entry = wxT("/"); PreviousDirs[0] = entry;	count=0;	// If there's only 1 path in array, this dud one, replace it with '/' (which SURELY still exists?).  Then display that.
startdir = entry; RefreshTree( entry );
}


void DirGenericDirCtrl::UpdateStatusbarInfo(wxString& selected)	// Writes filter info to statusbar(3), & the selection's name & data to statusbar(2)
{
if ( ! MyFrame::mainframe || ! MyFrame::mainframe->SelectionAvailable ) return;
														// First the filter info into the 4th statusbar pane
wxString filterinfo, start, filter = GetFilterString(); if ( filter.IsEmpty() )  filter = wxT("*");	// Translate "" into "*"
start = (GetShowHidden() ? wxT(" D H     ") : wxT(" D     "));						// Create a message string, D for dirs if applicable, H for Hidden ones
filterinfo.Printf(wxT("%s%s"), start.c_str(), filter.c_str() );							// Create a message string, adding the current filter
MyFrame::mainframe->SetStatusText( filterinfo, 3 ); 

if ( selected.IsEmpty() || partner==NULL ) 
	{ MyFrame::mainframe->SetStatusText( wxEmptyString, 2 ); return; }	// If no data, clear statusbar(2)

wxString text;
text.Printf(wxT("%s%s   %u Files, total size %s   %u Subdirectories"), _("Dir: "), selected.AfterLast( wxFILE_SEP_PATH ).c_str(),
											 (uint)((FileGenericDirCtrl*)partner)->NoOfFiles, ParseSize( ((FileGenericDirCtrl*)partner)->CumFilesize ).c_str(),
																																						 (uint)((FileGenericDirCtrl*)partner)->NoOfDirs );
MyFrame::mainframe->SetStatusText( text, 2 ); 
}

void MyGenericDirCtrl::OnShortcutCut(wxCommandEvent& WXUNUSED(event))								// Ctrl-X
{
MyGenericDirCtrl::filearray.Clear();								// We'll be adding to the clipboard, so start with a clean sheet
MyGenericDirCtrl::filecount = 0;

UnRedoManager::StartSuperCluster();							// Start a new UnRedoManager StartSuperCluster, to be continued by Paste
Delete( false, true );														// The 1st bool means Delete not Trash. The 2nd flags it's from Cut, so deleted files are stored in clipboard
UnRedoManager::EndCluster();										// Close the cluster (but Paste can reopen)
}


void DirGenericDirCtrl::NewFile(wxCommandEvent& WXUNUSED(event))	// Create a new Dir
{
wxString path, name;
																			// Where to put the new item
path = GetPath();																									// Get path of selected item
if ( path.IsEmpty() )  path = startdir;																		// If no selection, use "root"
if ( path.Last() != wxFILE_SEP_PATH )  path += wxFILE_SEP_PATH; 				// We almost certainly need to add a terminal '/'

FileData DestinationFD( path );
if ( ! DestinationFD.CanTHISUserWriteExec() )														// Make sure we have permission to write to the dir
	{	wxString msg;
		if  ( arcman != NULL && arcman->IsArchive() )  msg = _("I'm afraid you can't Create inside an archive.\nHowever you could create the new element outside, then Move it in.");
			else msg = _("I'm afraid you don't have permission to Create in this directory");
		wxMessageDialog dialog(this, msg, _("No Entry!"), wxOK | wxICON_ERROR );
		dialog.ShowModal(); return;
    }


wxDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("NewDir"));

int flag = 0;
do							// Do the rest in a loop, to give a 2nd chance if there's a name clash etc
	{	if ( dlg.ShowModal() != wxID_OK )		return;												// If user cancelled, abort
		
		name = ((wxTextCtrl*)dlg.FindWindow( wxT("FileDir_text") ))->GetValue();	// Get the desired name
		if ( name.IsEmpty() ) return;
		
		flag = MyGenericDirCtrl::OnNewItem( name, path, true );								// Do the rest in MyGenericDirCtrl, as shared with MyFiles version
	}
 while ( flag==wxID_YES );							// Loop if requested: means there was an error & we want another go
 
if ( !flag ) return;
 

wxArrayInt IDs; IDs.Add( GetId() );																			// Update the panes
MyFrame::mainframe->OnUpdateTrees( path, IDs );
UnRedoManager::StartCluster();																				// Start a new UnRedoManager cluster
UnRedoNewDirFile *UnRedoptr = new UnRedoNewDirFile( name, path, true, IDs );	// Make a new UnRedoNewDirFile
UnRedoManager::AddEntry( UnRedoptr );												 				// and store it for Undoing
UnRedoManager::EndCluster();
}

void DirGenericDirCtrl::ShowContextMenu(wxContextMenuEvent& event)
{
if ( ((MyTreeCtrl*)GetTreeCtrl())->QueryIgnoreRtUp() ) return;						// If we've been Rt-dragging & now stopped, we don't want a context menu

wxMenu menu(wxT(""));
menu.Append(SHCUT_CUT, _("Cut\tCtrl+X"));													// Note that, unless the Ctrl-X etc feature on the main menu, they won't display here!?
menu.Append(SHCUT_COPY, _("Copy\tCtrl+C"));
menu.Append(SHCUT_PASTE, _("Paste\tCtrl+V"));
menu.Append(SHCUT_SOFTLINK, _("Create Symlink\tAlt+Shift+L"));
menu.Append(SHCUT_HARDLINK, _("Create Hard Link\tCtrl+Shift+L"));
menu.AppendSeparator();

menu.Append(SHCUT_TRASH, _("Send Directory to Trashcan\tDEL"));			// A GTK bug prevents the display of Del on menus. Still works though
menu.Append(SHCUT_DELETE, _("Delete Directory\tSHIFT-DEL"));
menu.AppendSeparator();

menu.Append(SHCUT_REFRESH, _("Refresh this Display\tF5"));
menu.Append(SHCUT_RENAME, _("Rename Directory\tF2"));
menu.Append(SHCUT_DUP, _("Duplicate Directory"));
menu.Append(SHCUT_NEW, _("New Directory\tF10"));
menu.AppendSeparator();

menu.Append(SHCUT_FILTER, _("Filter display\tCtrl+Shift+F"));
if ( GetShowHidden() )
	menu.Append(SHCUT_TOGGLEHIDDEN, _("Hide hidden dirs and files\tCtrl+H"));
 else
 	menu.Append(SHCUT_TOGGLEHIDDEN, _("Show hidden dirs and files\tCtrl+H"));
menu.AppendSeparator();

menu.Append(SHCUT_SPLITPANE_VERTICAL,_("Split Panes Vertically"));
menu.Append(SHCUT_SPLITPANE_HORIZONTAL,_("Split Panes Horizontally"));
menu.Append(SHCUT_SPLITPANE_UNSPLIT, _("Unsplit Panes "));
menu.AppendSeparator();
menu.Append(SHCUT_REPLICATE, _("Replicate path in opposite Pane"));
menu.Append(SHCUT_SWAPPANES, _("Swap path with opposite Pane"));
menu.AppendSeparator();

menu.Append(SHCUT_PROPERTIES, _("Properties"));

wxPoint pt = event.GetPosition();
ScreenToClient(&pt.x, &pt.y);

PopupMenu(&menu, pt.x, pt.y);
}


void MyGenericDirCtrl::OnSplitpaneVertical(wxCommandEvent& WXUNUSED(event))			// Vertically splits the panes of the current tab
{
ParentTab->PerformSplit( vertical );
}

void MyGenericDirCtrl::OnSplitpaneHorizontal(wxCommandEvent& WXUNUSED(event))	// Horizontally splits the panes of the current tab
{
ParentTab->PerformSplit( horizontal );
}

void MyGenericDirCtrl::OnSplitpaneUnsplit(wxCommandEvent& WXUNUSED(event))			// Unsplits the panes of the current tab
{
ParentTab->PerformSplit( unsplit );
}


void MyGenericDirCtrl::OnShortcutTrash(wxCommandEvent& WXUNUSED(event))				// Del
{
UnRedoManager::StartCluster();													// Start a new UnRedoManager cluster
Delete( true );
UnRedoManager::EndCluster();													// Close the cluster
}


void MyGenericDirCtrl::OnShortcutDel(wxCommandEvent& WXUNUSED(event))					// Sh-Del
{
UnRedoManager::StartCluster();													// Start a new UnRedoManager cluster
Delete( false );
UnRedoManager::EndCluster();													// Close the cluster
}

void MyGenericDirCtrl::OnDnDMove()						// Move by DnD
{
wxString DestPath;
enum DupRen WhattodoifClash = Unknown, WhattodoifCantRead = Unknown;		// These get passed to CheckDupRen.  May become SkipAll, OverwriteAll etc, but start with Unknown

										// The item(s) 2b moved are in DnD clipboard
if ( ! MyGenericDirCtrl::DnDfilecount ) return;							// If no files selected, abort
if ( MyGenericDirCtrl::DnDDestfilepath == wxEmptyString ) return;	// If no destination selected, abort

wxString path = MyGenericDirCtrl::DnDDestfilepath;

										// We need to check whether we're moving into or from within an archive: if so things are different
MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane();	// Active is the origin MyGenericDirCtrl, 'this' is the destination one
if ( active != NULL && active->arcman != NULL && active->arcman->IsArchive() ) // The origin pane is an archive
	{	if ( active->arcman->GetArc()->MovePaste( active, this, MyGenericDirCtrl::DnDfilearray, path ) )
			DoBriefLogStatus( MyGenericDirCtrl::DnDfilearray.GetCount(), wxEmptyString,  _("moved") ); // Use GetCount() as some items may have been skipped & removed from the array
		return;
	}
if ( arcman != NULL && arcman->IsArchive() ) 														// The destination pane is an archive
	{	if ( arcman->GetArc()->MovePaste( active, this, MyGenericDirCtrl::DnDfilearray, path ) )
			DoBriefLogStatus( DnDfilecount, wxEmptyString,  _("moved") );
		return;
	}
										// Make sure the destination path is a dir with a terminal /  It's probably missing it
wxFileName Destination( path );
FileData fd( path );

if ( fd.IsDir() || fd.IsSymlinktargetADir(true) )								// If the overall path is indeed a dir, plonk a / on the end of path, for the future
	{ if ( ! Destination.GetFullName().IsEmpty() )  path += Destination.GetPathSeparator(); }
 else 																							// If it's a FILE, remove the filename, leaving the path for pasting to
	{ Destination.SetFullName( wxEmptyString ); path = Destination.GetFullPath(); }

FileData DestinationFD( path );
if ( ! DestinationFD.CanTHISUserWriteExec() )							// Make sure we have permission to write to the destination dir
	{	wxMessageDialog dialog(this, _("I'm afraid you don't have Permission to write to this Directory"), _("No Entry!"), wxOK | wxICON_ERROR );
		dialog.ShowModal(); return;
	}

wxArrayInt IDs; IDs.Add( MyGenericDirCtrl::DnDFromID );		// Make an int array, & store the ID of origin pane
IDs.Add( GetId() );																		// Add the destination ID

FileData OriginFileFD( MyGenericDirCtrl::DnDfilearray[0] );
if ( ! OriginFileFD.CanTHISUserMove_By_RenamingThusly( path ) )	// Make sure we have permissions for the origin dir to Rename its files   i.e. the new path is still on the same device/partition
	if ( ! ( OriginFileFD.CanTHISUserMove()  && CanFoldertreeBeEmptied( MyGenericDirCtrl::DnDfilearray[0] )==CanBeEmptied ) )	//   or alternatively, that we will be allowed to do a true Move ie copy + delete the original
		{	if ( OriginFileFD.CanTHISUserPotentiallyCopy() )			// See if we have parent dir exec (even though not write) permission
				{	wxMessageDialog dialog(this, _("I'm afraid you don't have the right permissions to make this Move.\nDo you want to try to Copy instead?"),
																																										_("No Exit!"), wxYES_NO | wxICON_QUESTION );
					if ( dialog.ShowModal() != wxID_YES )		return;
					 else return OnPaste( true, path );						// Use Paste instead
				}
			 else 
				{	wxMessageDialog dialog(this, _("I'm afraid you don't have the right permissions to make this Move"), 	_("No Exit!") ); return; }
		}

int successes = 0;
for ( size_t n=0; n < DnDfilecount; ++n )									// For every entry in the clipboard
	{	wxString DestFilename, DestOriginalname;
		bool ItsADir;
		
		FileData FromFD( MyGenericDirCtrl::DnDfilearray[n] );		// Make this File/Dir being dragged a FileData

														// We need to check we're not about to do something illegal or immoral
		CheckDupRen CheckIt(this, MyGenericDirCtrl::DnDfilearray[n], path);	// Make an instance of the class that tests for this
		CheckIt.IsMultiple = ( (n+1) < DnDfilecount );						// This flags whether there'll be more items after this one, so use different dialogs
		CheckIt.WhattodoifClash = WhattodoifClash; CheckIt.WhattodoifCantRead = WhattodoifCantRead;	// Set CheckIt's variables to match any previous choice
		if ( ! CheckIt.CheckForAccess() ) 										// This checks for Read permission failure, and sets a SkipAll flag if appropriate
			{ WhattodoifCantRead = CheckIt.WhattodoifCantRead; continue; }
		if ( CheckIt.CheckForIncest()  											// Check we're not pasting onto a descendant.  False means OK
				||  !CheckIt.CheckForPreExistence() )							//  or onto ourselves or a namesake.			   True means OK
			{ WhattodoifClash = CheckIt.WhattodoifClash;				// Store any new choice in local variable.  This catches SkipAll & Cancel, which return false
				if ( WhattodoifClash==Cancel ) 
					{ n = DnDfilecount; break; }									// Cancel this item & the rest of the loop by increasing n
				 else continue; 															// Skip this item but continue with loop
			}
		 else 					// If there was a clash & subsequent Rename, CheckDupRen will have changed the filename or the last segment of the path
			{	DestPath = CheckIt.finalpath;										// This will be either the original path without the filename/last dir-segment, or the altered version
				DestFilename = CheckIt.finalbit;									// Ditto, but the filename/last-segment
				WhattodoifClash = CheckIt.WhattodoifClash;				// Store any new choice in local variable
				ItsADir = CheckIt.ItsADir;											// CheckIt already tested whether it's a file or a dir.  Piggyback onto this, it saves constant rechecking
			}

		if ( ItsADir && MyGenericDirCtrl::DnDfilearray[n].Last() != wxFILE_SEP_PATH )	// If we're moving a dir, make sure it ends in a /
			{	MyGenericDirCtrl::DnDfilearray[n] += wxFILE_SEP_PATH;							// Also, store the last segment of the path, before any rename
				DestOriginalname = (MyGenericDirCtrl::DnDfilearray[n].BeforeLast(wxFILE_SEP_PATH)).AfterLast(wxFILE_SEP_PATH);	
			}
		 else	DestOriginalname = MyGenericDirCtrl::DnDfilearray[n].AfterLast(wxFILE_SEP_PATH);	// If it's a file, store the original filename, before any rename
			
		wxFileName From( MyGenericDirCtrl::DnDfilearray[n] );	// Make this File/Dir being dragged into a wxFileName
		wxFileName To( DestPath );												// Ditto the To path  Must be done in each loop, as Move alters its contents

		if ( ! Move( &From, &To, DestFilename ) )	continue;		// Subcontract the actual Move
		
																					// If we're here, it worked. 
		if ( FromFD.IsDir() )
						From.RemoveDir( From.GetDirCount() - 1 );		// If we just dragged a dir, readjust wxFileName to take into account the removal. 
		 else		From.SetFullName( wxEmptyString ); 				// If a file/Symlink, remove its name
					// NB The Move already altered Destination, appending to it the pasted subdir. Fortunately, this is just what we want for Undoing

		wxString frompath = From.GetFullPath(), topath = To.GetFullPath();
		UnRedoMove *UnRedoptr = new UnRedoMove( frompath, IDs, topath, DestFilename, DestOriginalname, ItsADir );	// Make a new unredoMove
		UnRedoManager::AddEntry( UnRedoptr );		 					// and store it for Undoing			
				 					// Sort out the treectrl.  UpdateTrees does just that
		MyFrame::mainframe->OnUpdateTrees( From.GetPath(), IDs );
		MyFrame::mainframe->OnUpdateTrees( DestPath, IDs );
		++successes;
	}

DoBriefLogStatus( successes, wxEmptyString,  _("moved") );
}

void MyGenericDirCtrl::OnShortcutHardLink(wxCommandEvent& WXUNUSED(event))
{
UnRedoManager::StartCluster();							// Start a new UnRedoManager cluster
OnLink( false, myDragHardLink );						// I'm recycling myDragHardLink, even though this isn't a Drag, as flagged by the 'false'
UnRedoManager::EndCluster();							// Close the cluster
}

void MyGenericDirCtrl::OnShortcutSoftLink(wxCommandEvent& WXUNUSED(event))
{
UnRedoManager::StartCluster();
OnLink( false, myDragSoftLink );						// I'm recycling myDragSoftLink, even though this isn't a Drag, as flagged by the 'false'
UnRedoManager::EndCluster();
}

void MyGenericDirCtrl::OnLink( bool FromDnD, int linktype )			// Create hard or soft link
{
size_t count;
wxArrayString filearray;
wxArrayInt IDs;
wxString path, ext, linkname; bool samedir; int nametype; wxRadioBox* LinktypeRadio;
												// We may be using Ctrl-V or Dnd, so get data into local vars accordingly
if ( FromDnD )
	{	count = MyGenericDirCtrl::DnDfilecount;
		filearray = MyGenericDirCtrl::DnDfilearray;
		IDs.Add( GetId() );										// Store the ID of the destination pane
		path = MyGenericDirCtrl::DnDDestfilepath;
	}
 else
 	{	count = MyGenericDirCtrl::filecount;
		filearray = MyGenericDirCtrl::filearray;
		IDs.Add( GetId() );
		path = GetPath();
		if ( path.IsEmpty() && fileview==ISRIGHT )	// If we're in a fileview pane, but not hovering over something, 
				path = startdir;									//    get the root path for the pane
				
		FileData fd( path );
		if ( fd.IsSymlink() )										// If we're being asked to Paste onto a symlink, see if its target is a dir
			{	wxString newpath = fd.GetUltimateDestination();
				FileData nfd( newpath ); 
				if ( nfd.IsDir() )  path = newpath;		// Yes, it's a symlink-to-dir, so we want to replace path with the symlink target
			}																// NB If the target isn't a dir, we don't want to deref the symlink at all; just paste into its parent dir as usual
	
	}

if ( ! count )  return; 											// Nothing in the "Clipboard"
if ( path==wxEmptyString )  return; 					// Nowhere to put it (not very likely, but nevertheless)

wxFileName Destination( path );							// I'm making sure the destination path is a dir with a terminal /  It's probably missing it
FileData fd( path );
if ( fd.IsDir() && ! Destination.GetFullName().IsEmpty() )	// If the overall path is indeed a dir
		path += wxFILE_SEP_PATH;										//   plonk a / on the end of path, for the future
	else 
		{	Destination.SetFullName( wxEmptyString );			// If it IS a file, remove the filename, leaving the path for pasting to
			path = Destination.GetFullPath();
		}

FileData fd1( path );											// Make a new FileData as path may have been altered above
if ( ! fd1.CanTHISUserWriteExec() )					// Ensure we have permission to write to the destination dir
	{	wxString msg;
		if  ( arcman != NULL && arcman->IsArchive() )  msg = _("I'm afraid you can't create links inside an archive.\nHowever you could create the new link outside, then Move it in.");
			else msg = _("I'm afraid you don't have permission to create links in this directory");
		wxMessageDialog dialog(this, msg, _("No Entry!"), wxOK | wxICON_ERROR );
		dialog.ShowModal(); return;
    }
					// We need to be able to compare the source path with the dest path, to ensure we don't copy onto ourselves without changing name
wxString sourcepath = filearray[0].BeforeLast(wxFILE_SEP_PATH);
while ( filearray[0].Right(1) == wxFILE_SEP_PATH  && filearray[0].Len() > 1 ) sourcepath.RemoveLast();
samedir = ( sourcepath == path.BeforeLast(wxFILE_SEP_PATH) );		// See if they're the same (we just added a terminal '/' to path, so remove if first)

MakeLinkDlg dlg; wxXmlResource::Get()->LoadDialog(&dlg, this, wxT("MakeLinkDlg"));
dlg.init( count > 1, samedir );																	// If count<2 it doesn't make sense to offer to apply choices to other files.  If samedir, compel a name-change

int successes = 0;
for ( size_t c=0; c < count; ++c )							// For every path in the clipboard to be linked
	{	if ( ! ( dlg.ApplyToAll->IsEnabled() && dlg.ApplyToAll->GetValue() ) )	// If this isn't the 1st time thru the loop, see if we've been told to ApplyToAll.  If so, skip to the 'Do it' section
			{	wxString msg; msg.Printf( _("Create a new %s Link from:"), (linktype==myDragHardLink ? _("Hard") : _("Soft")) );
				((wxStaticText*)dlg.FindWindow( wxT("link_message") ))->SetLabel( msg );		// Adjust the label to state link type
				LinktypeRadio = (wxRadioBox*)dlg.FindWindow( wxT("link_radiobox") );
				LinktypeRadio->SetSelection( (linktype==myDragSoftLink) );
				wxString str;int x, y; 
				str = filearray[c];	dlg.frompath->GetTextExtent(str, &x, &y ); // 2.4.2 is OK, but without this manoevre, later versions don't centre properly
				if ( x > 300 )																			// 300 is the intended max width of the box. Truncate if greater
						{ dlg.frompath->SetSize( 300, 25);dlg.frompath->SetLabel( TruncatePathtoFitBox( str, dlg.frompath ) );}	// Load the FROM filepath
				  else	dlg.frompath->SetLabel( str );										// No truncating needed this time
					
				str = path;	dlg.topath->GetTextExtent(str, &x, &y );				// Ditto TO
				if ( x > 300 )	{ dlg.topath->SetSize( 300, 25);dlg.topath->SetLabel( TruncatePathtoFitBox( str, dlg.topath ) );}	// Load the FROM filepath
				  else	dlg.topath->SetLabel( str );											// No truncating needed this time

				if ( dlg.lastradiostation == 2 )													// If we're likely to be making the link from the filename, insert the filename
					dlg.LinkFilename->SetValue( filearray[c].AfterLast(wxFILE_SEP_PATH) );
				dlg.GetSizer()->Layout();
				
				int result;
				do
					{	result = false;
						int ans = dlg.ShowModal();
						if ( ans == wxID_CANCEL ) return;						// Cancel everything else
						if ( ans != wxID_OK )			continue; 					// If user pressed Skip, abort this iteration of the main loop
						result = true;
						
						if  ( ((wxRadioButton*)dlg.FindWindow( wxT("NameSame") ))->GetValue() ) nametype = 0;	// How do we want to make a name for the link:  reuse the original, add an ext, or bespoke?
						  else if  ( ((wxRadioButton*)dlg.FindWindow( wxT("NameExt") ))->GetValue() ) nametype = 1;
							else nametype = 2;

						if ( nametype == 1 )   												// Add an ext	
							{	ext = dlg.LinkExt->GetValue();
								if ( ext.IsEmpty() )
									{	wxMessageBox( _("Please provide an extension to append") );
										result = 2; continue;
									}
								if ( ext.at(0) != wxT('.') ) ext = wxT('.') +ext;	// Ensure the ext is an ext!
							}
						if ( nametype == 2 &&	dlg.LinkFilename->GetValue().IsEmpty() )
								{	wxMessageBox( _("Please provide a name to call the link") );
									result = 2; continue;
								}
						if ( nametype == 2 &&	dlg.LinkFilename->GetValue() == filearray[c].AfterLast(wxFILE_SEP_PATH) )
								{	wxMessageBox( _("Please provide a different name for the link") );
									result = 2; continue;
								}		
					} while ( result == 2 );												// Redo the dialog if there was insufficient data
				if ( !result ) continue;														// Cancel this iteration
			}
			
		if ( nametype == 0 )  linkname = filearray[c].AfterLast(wxFILE_SEP_PATH); 						// Reuse the filename
		 else if ( nametype == 1 )		linkname = filearray[c].AfterLast(wxFILE_SEP_PATH) + ext;	//  or add an ext
		 else if ( nametype == 2 )		linkname = dlg.LinkFilename->GetValue();								//    or use a fresh name
		 
		wxString linkfilepath = path + linkname;									// Create the new filepath

		bool success; bool linkage = LinktypeRadio->GetSelection();		// Reread the radiobox to see if we had 2nd thoughts about the type of link
		if ( linkage )	
			{	int rel = dlg.MakeRelative->IsChecked();								// Do we want to make a relative- or absolute-path symlink?
				success = CreateSymlinkWithParameter( filearray[c], linkfilepath, (enum changerelative)rel );	// Create a new symlink
				if ( ! success )	continue;
				
				UnRedoLink *UnRedoptr = new UnRedoLink( filearray[c], IDs, linkfilepath, linkage, (enum changerelative)rel );
				UnRedoManager::AddEntry( UnRedoptr );
			}
		  else	
			{ 	success = CreateHardlink( filearray[c], linkfilepath );																	// 	or hardlink
				if ( ! success )	continue;
				
				UnRedoLink *UnRedoptr = new UnRedoLink( filearray[c], IDs, linkfilepath, linkage );
				UnRedoManager::AddEntry( UnRedoptr );
			}
			
		++successes;

		
		MyFrame::mainframe->OnUpdateTrees( linkfilepath, IDs );
		if ( FromDnD && fileview==ISLEFT )	partner->ReCreateTreeFromSelection();	// Symlinks made into a dirview will otherwise not update the fileview
	}

DoBriefLogStatus( successes, wxEmptyString,  _("linked") );
}

void MyGenericDirCtrl::Delete( bool trash, bool fromCut /*=false*/ )	// Does either Delete or Trash, depending on 1st bool
{
wxString msg, path, DestFilename;
wxArrayString paths;
bool ItsADir;
wxString action;
if ( fromCut )  action = _("cut");
 else action = (trash ?  _("trashed") : _("deleted"));
		
			// We need to get the highlit item(s) to be deleted.  If wxTR_MULTIPLE there may be >1
if ( GetTreeCtrl()->HasFlag( wxTR_MULTIPLE ) )     GetMultiplePaths( paths );	// Fill array with multiple selections, if style permits
	else  	paths.Add( GetPath() );														// If not multiple selection style, get the one&only selection

size_t count = paths.GetCount(); if ( !count ) return;						// If no selection, abort

bool IsArchive = (arcman != NULL && arcman->IsArchive());			// Archives are very different
if ( ! IsArchive )
	{	FileData fd0( paths[0] );																// Make a FileData to check for delete permission for the parent dir
		if ( ! fd0.CanTHISUserMove() )
			{	wxMessageDialog dialog(this, _("I'm afraid you don't have permission to Delete from this directory"), _("No Entry!"), wxOK | wxICON_ERROR );
				dialog.ShowModal(); return;
			}
	}

if ( trash )	msg = _("Discard to Trash: "); 
 else msg = _("Delete: ");

if ( ((ASK_ON_DELETE && !trash) || (ASK_ON_TRASH && trash))	// If it's preferred not to destroy the filesystem accidentally
																			&& !fromCut )			//   & we're deleting, not Cutting
	{	if ( count < 10 )
			for ( size_t n=0; n < count; n++ )											// Get the path(s) into path, separated by newlines
				path +=  wxT("\n") + paths[n];			
		 else																							// If there are too many items, the MessageDialog expands off the page! So prune
		 	{	path.Printf( _("%u items, from: "), count );
				for ( size_t n=0; n < 3; n++ )												// Get first 3 paths into path, separated by newlines
					path +=  wxT("\n") + paths[n];
				path +=  _("\n\n          To:\n");
				for ( size_t n = count-3; n < count; n++ )							// & finish with the last 3 paths
					path +=  wxT("\n") + paths[n];
			}
		wxMessageDialog ask(this,msg+path,wxString(_("Are you SURE?")), wxYES_NO | wxICON_QUESTION  );
		int ans = ask.ShowModal();
		if ( ans != wxID_YES ) return;													// The user wanted to bail out
	}

wxFileName trashdirbase;																	// Create a unique subdir in trashdir, using current date/time
if ( ! DirectoryForDeletions::GetUptothemomentDirname( trashdirbase, trash ? trashcan : delcan ) )
	{ wxMessageBox( _("For some reason, trying to create a dir to receive the deletion failed.  Sorry!")); return; }

#if wxVERSION_NUMBER >= 2504
if ( IsArchive )						
	{	wxArrayString destinations, unused; wxString trashpath( trashdirbase.GetPath() );
		if ( ! arcman->GetArc()->DoExtract( paths, trashdirbase.GetPath(), destinations, true ) )	// First extract the files to the bin. The true means don't unredo: it'll happen elsewhere
			{ wxMessageDialog dialog( this, _("Sorry, Deletion failed"), _("Oops!"), wxOK | wxICON_ERROR ); dialog.ShowModal(); return; }
		arcman->GetArc()->Alter( paths, arc_remove, unused );											// Then remove from the archive
		wxArrayInt IDs; IDs.Add( GetId() );
		MyFrame::mainframe->OnUpdateTrees( paths[0], IDs );
		UnRedoMoveToFromArchive* UnRedoptr = new UnRedoMoveToFromArchive( paths, destinations, paths[0], trashpath, IDs, this, this, amt_from  );
		UnRedoManager::AddEntry( UnRedoptr );
		if ( fromCut )																				// If we're Cutting, we now need to fill the 'clipboard' ready for any Paste
			{	wxDir d( trashpath ); if ( ! d.IsOpened() ) return;
				wxString filename; bool cont = d.GetFirst( &filename );
       			while ( cont )	{ MyGenericDirCtrl::filearray.Add( trashpath+wxFILE_SEP_PATH+filename ); cont = d.GetNext( &filename ); }
				MyGenericDirCtrl::filecount = MyGenericDirCtrl::filearray.GetCount(); 
				MyGenericDirCtrl::Originpane = this;
			}
		DoBriefLogStatus( count, wxEmptyString, action );
		return;
	}
#endif // #if wxVERSION_NUMBER >= 2504

size_t successes=0, cantdel=0, cantdelsub=0, invalid=0;					// so that for multiple deletions, we can if necessary report on partial success

for ( size_t n=0; n < count; n++ )														// For every path in the array
	{	wxString original;
		wxFileName trashdir;
		
		if ( count > 1 )																				// If we're doing a multiple del, create a unique subdir for each item (in case of del.ing both ./foo & ./bar/foo)
			{	wxString subdirname = trashdirbase.GetFullPath() + CreateSubgroupName( n, count );			// Create a unique subgroup name
				trashdir.Mkdir( subdirname );												// Create the subdir to which to delete
				trashdir.AssignDir( subdirname );
			}	
		 else trashdir = trashdirbase;														// If only one item to delete, use the base dir

		FileData stat( paths[n] ); ItsADir = stat.IsDir();
		if ( ItsADir )
			{	enum emptyable ans = CanFoldertreeBeEmptied( paths[n] );// Check this dir and any downstream can be deleted from
				if ( ans != CanBeEmptied )
					{	switch ( ans )
							{	case CannotBeEmptied:				++cantdel; break;
								case SubdirCannotBeEmptied:	++cantdelsub; break;
								 default: 									++invalid; break;
							}
						continue;																	// Try another filepath
					}
			
				if ( paths[n].Last() != wxFILE_SEP_PATH ) paths[n] << wxFILE_SEP_PATH;						// If we're deleting a dir, make sure it ends in a /
				DestFilename = (paths[n].BeforeLast(wxFILE_SEP_PATH)).AfterLast(wxFILE_SEP_PATH);	// We need the last segment of the path
			}
		 else	DestFilename = paths[n].AfterLast(wxFILE_SEP_PATH);	// If it's a file, store the filename
		 	
		wxFileName fn( paths[n] );															// Turn filepath into the class that does clever things to the file system
		
		if ( ! Move(&fn, &trashdir, DestFilename) )	break;						// Do the actual "deletion"

		++successes;						// If we're here, it worked.  Readjust wxFileName to take into account the deletion. 
		if ( ItsADir )																					// If it's a dir,
			{ original = fn.GetPath();															// Save current path
				fn.RemoveDir( fn.GetDirCount() - 1 );									// then readjust wxFileName to take into account the deletion.
			}
		 else
		 	 fn.SetFullName(wxEmptyString); 											// If it's a file, just kill the filename, leaving the path unchanged
			
		path = fn.GetPath();																	//  Either way, load path with truncated version
	
											// NB The Move already altered To, appending to it the pasted subdir. Fortunately, this is just what we want for Undoing
		wxArrayInt IDs;																			// Make an int array to store the ID of the origin pane
		IDs.Add( GetId() );

		wxString frompath = fn.GetFullPath(), topath = trashdir.GetFullPath();
		UnRedoMove *UnRedoptr = new UnRedoMove( frompath, IDs, topath, DestFilename, DestFilename, ItsADir, true );	// Make a new unredoMove
		UnRedoManager::AddEntry( UnRedoptr );		 							// (The 'true' in the line above signifies "From Delete", so UnRedo knows not to UpdateTrees for trashdir)

		if ( fromCut )
			{	MyGenericDirCtrl::filearray.Add( topath );							// Store the address of the deleted item in the "clipboard"
				++MyGenericDirCtrl::filecount;											// Inc its count
			}
																											// Finally we have to sort out the treectrl
		if ( fileview==ISLEFT && GetPath()==stat.GetFilepath() )			// If we've just deleted the selected dir
							SetPath( path );														//    we need to SetPath to elsewhere, otherwise the fileview continues to show the contents of the deleted dir!

		if ( (ItsADir && startdir==original) || (startdir==original.BeforeLast(wxFILE_SEP_PATH)) )	// Make sure we're not deleting 'root'
					{	while ( path.Right(1) == wxFILE_SEP_PATH && path.Len() > 1 ) path.RemoveLast();		  // Cope with "/foo///" and also "///"
						MyFrame::mainframe->OnUpdateTrees( original, IDs, path );// If we are, use the alt version of updatetrees to replace startdir with its parent (path)
					}
		  else MyFrame::mainframe->OnUpdateTrees( path, IDs );		// Otherwise the standard version
	}

if ( cantdel || cantdelsub || invalid )														// Were there any failures above?  If so, do an apologetic dialog
	{	wxString msg, msg2, msg3, msg4;
		if ( count == 1 )
			{	if ( cantdel ) msg2 =  _(" You don't have permission to Delete from this directory.\nYou could try deleting piecemeal.");
					else if ( cantdelsub ) msg2 =   _(" You don't have permission to Delete from a subdirectory of this directory.\nYou could try deleting piecemeal.");
						else if ( invalid )  msg2 =   _(" The filepath was invalid.");
			}
		else																								// More complex message if  >1 item
			{	if ( cantdel ) msg2.Printf(wxT("\n%s%u%s"),_("For "), (uint)cantdel, _(" of the items, you don't have permission to Delete from this directory.") );	
				if ( cantdelsub ) msg3.Printf(wxT("\n%s%u%s"),_("For "), (uint)cantdelsub, _(" of the items, you don't have permission to Delete from a subdirectory of this directory.") );
				if ( invalid )  msg4.Printf(wxT("\n%s%u%s"),_("For "), (uint)invalid, _(" of the items,  the filepath was invalid.") );
				msg2 =  msg2 + msg3 + msg4;											// Concatanate to cater for multiple failure modalities
				if ( ! msg2.IsEmpty() )  msg2 += _(" \nYou could try deleting piecemeal.");
			}
		if ( ! successes )																			// Abject failure
			{	if ( count == 1 )	msg = _("I'm afraid the item couldn't be deleted.");		// Only 1 item so different grammar
					else					msg.Printf( wxT("%s%u%s"),_("I'm afraid the "), (uint)count, _(" items could not be deleted.") );
			}
		 else																							// If there was at least one success
				msg.Printf( wxT("%s%u%s%u%s"),_("I'm afraid only "), (uint)successes, _(" of the "), (uint)count, _(" items could be deleted.") );

		wxMessageDialog dialog( this, msg + msg2, _("Oops!"), wxOK | wxICON_ERROR ); dialog.ShowModal();
	}

if ( ! successes ) { BriefLogStatus bls( fromCut ?  _("Cut failed") : _("Deletion failed") ); return; }

DoBriefLogStatus( count, wxEmptyString, action );
}

//static
bool MyGenericDirCtrl::ReallyDelete( wxFileName* PathName )	// Deletes files, dirs+contents.  Static as used by UnRedoManager too
{
wxString msg;
				// Getting paths from the wxTreeCtrl  has a disadvantage:  if a dir is selected, it has no terminal /   The same's true of wxDir.
				// As a result wxFileName assumed that the name is path+filename, and this screws up the subsequent procedures.
				// So check whether it's actually a dir, and if so to append the "filename" to the path.
				// There's also the minor problem of symlinks, which may target a dir!  So use FileData
class FileData stat( PathName->GetFullPath() );
if ( stat.IsDir() )																					// If it's really a dir
	{	if ( ! PathName->GetFullName().IsEmpty() )								// and providing there IS a "filename"
				PathName->AppendDir( PathName->GetFullName() );  PathName->SetFullName( wxEmptyString ); // make it into the real path
	}
 else
	{	if ( stat.IsSymlink() )
			{	wxBusyCursor busy;
				if ( ! wxRemoveFile( stat.GetFilepath() ) )
					{ wxMessageBox(_("Deletion Failed!?!")); return false; }	// Shouldn't fail, but exit ungracefully if somehow it does eg due to permissions
			}			
		 else 													// It must be a file
			{	if ( ! stat.IsValid() ) 															// Check it exists.  If not, exit ungracefully
					{	msg = _("? Never heard of it!"); wxMessageBox(PathName->GetFullName()+msg); return false; }
				wxBusyCursor busy;
				if ( ! wxRemoveFile( PathName->GetFullPath() ) )				// It's alive, so kill it
					{ wxMessageBox(_("Deletion Failed!?!")); return false; }	// Shouldn't fail, but exit gracelessly if somehow it does eg due to permissions
			}
		return true;
	}

																	// If we're here, it's a dir
wxBusyCursor busy;

wxDir dir( PathName->GetPath() );														// Use helper class to deal sequentially with each child file & subdir
if ( !dir.IsOpened() )        return false;												// Oops

wxString filename;

while ( dir.GetFirst( &filename ) )														// Go thru the dir, committing infanticide 1 child @ a time
	{
		wxFileName child( PathName->GetPath(), filename );				// Make a new wxFileName for child
		if ( ! ReallyDelete( &child ) )														// Slaughter it by recursion, whether it's a dir or file
				return false;																		// If this fails, bug out
	}
	
if ( ! PathName->Rmdir() )																	// Finally, kill the dir itself
	{ wxMessageBox(_("Directory deletion Failed!?!")); return false; }

return true; 
}

//static 
bool MyGenericDirCtrl::Move( wxFileName* From, wxFileName* To, wxString ToName )	// Moves files, dirs+contents.  Static as used by UnRedo too
{
wxLogNull log;
wxFileName OldTo( *To );																// Store the unamended destination, in case we need it later

FileData from( From->GetFullPath() );
if ( ! from.IsValid() )
	{ wxMessageBox(_("The File seems not to exist!?!")); return false; }	

if ( from.IsDir() )													// If we're Moving a dir, we need to add it to the path
	{	wxString fullpath( To->GetPath() ); if ( fullpath.Last() != wxFILE_SEP_PATH ) fullpath << wxFILE_SEP_PATH;
		if ( !ToName.IsEmpty() && ToName.Last() != wxFILE_SEP_PATH ) ToName << wxFILE_SEP_PATH;
		
		if ( ! (ToName.Left( ToName.Len()-1 )).BeforeLast(wxFILE_SEP_PATH).IsEmpty() ) // See if there's an intermediate dir that we have to make ie foo/bar/ instead of just bar/
			{	wxFileName temp( fullpath + ToName );							// If so, create a wxFileName, truncate it and mkdir
				temp.RemoveDir( temp.GetDirCount() - 1 );
				temp.Mkdir(0777, wxPATH_MKDIR_FULL);
			}
		
		To->Assign( fullpath + ToName );											// Use Assign here, rather than AppendDir, in case ToName is foo/bar/
	}
 else																	// If it's not a dir, make it the filename
	{	wxString ToNamePath = ToName.BeforeLast(wxFILE_SEP_PATH), tooname(ToName);
		if ( ! ToNamePath.IsEmpty() ) tooname = tooname.AfterLast(wxFILE_SEP_PATH);
		wxString fullpath( To->GetPath() ); if ( fullpath.Last() != wxFILE_SEP_PATH ) fullpath << wxFILE_SEP_PATH;
		if ( !ToNamePath.IsEmpty() && ToNamePath.Last() != wxFILE_SEP_PATH ) ToNamePath << wxFILE_SEP_PATH;
		ToNamePath = fullpath + ToNamePath;
		To->Assign( ToNamePath, tooname );									// Use Assign here, rather than SetFullName(), in case ToName is foo/bar
	}

wxBusyCursor busy;

	// There are 2 ways to Move a file or dir:  the easy way is just to rename it.  However this fails if the destination filepath is on a different partition.
	// It also fails if we're moving a symlink with a relative target (which we don't want it to retain) so use the second method for that (I've altered Paste() to cope)
if ( RETAIN_REL_TARGET && ! ContainsRelativeSymlinks( from.GetFilepath() ) )
	if ( wxRename( From->GetFullPath(), To->GetFullPath() ) == 0 )   return true;

	// If we're here, it was either a relative symlink or renaming failed (we could look at errno to see if it's EXDEV if we're inquisitive).  To try the alternative, Copy/Delete, method
if ( ! Paste( From, &OldTo, ToName, from.IsDir() ) ) return false;
return ReallyDelete( From );
}

//static 
bool  MyGenericDirCtrl::Paste( wxFileName* From, wxFileName* To, wxString& ToName, bool ItsADir )	// Copies files, dirs+contents.  Static as used by UnRedo too
{
wxLogNull log;

FileData stat( From->GetFullPath() );

if ( ! stat.IsDir() )										// Check first to see if From isn't a dir.  If so, it's easy: just it, no progeny
	{	wxString ToNamePath = ToName.BeforeLast(wxFILE_SEP_PATH), tooname(ToName); // However these lines are in case ToName is foo/bar, and foo/ doesn't exist
		if ( ! ToNamePath.IsEmpty() ) 
			{	tooname = tooname.AfterLast(wxFILE_SEP_PATH);
				wxString fullpath( To->GetPath() ); if ( fullpath.Last() != wxFILE_SEP_PATH ) fullpath << wxFILE_SEP_PATH;
				ToNamePath = fullpath + ToNamePath; if ( ToNamePath.Last() != wxFILE_SEP_PATH ) ToNamePath << wxFILE_SEP_PATH;
				To->Assign( ToNamePath );
				if ( ! To->DirExists() )
					if ( ! To->Mkdir( 0777, wxPATH_MKDIR_FULL ) )	return false;						// wxPATH_MKDIR_FULL means intermediate dirs are created too
			}
		
		To->SetFullName( tooname );																							// Add the filename to create the desired dest. path/name

		if ( stat.IsSymlink() )																										// If it's a symlink, make a new one in destination dir
			{	wxString symlinktarget = stat.GetSymlinkDestination( true );
				if ( RETAIN_REL_TARGET && symlinktarget.GetChar(0) != wxFILE_SEP_PATH )	// If it's a relative symlink (& we want to) we need to be clever to keep the same target
					{	wxFileName fn( symlinktarget ); fn.MakeAbsolute( From->GetPath() );			// So first make the target absolute, relative to the original symlink,
						fn.MakeRelativeTo( To->GetPath() );	symlinktarget = fn.GetFullPath();		//  then relative to the new one
					}				
				if ( ! CreateSymlink( symlinktarget, To->GetFullPath() ) )  return false;
			}
		 else
		  if ( stat.IsRegularFile() )																								// If it's a normal file, use built-in wxCopyFile
		 	{	wxBusyCursor busy;
				if ( ! wxCopyFile( From->GetFullPath(), To->GetFullPath(), false ) )						// Copy it
					return false;
			}
		 else
		 	{	wxBusyCursor busy;																								// Otherwise it's an oddity like a fifo, which makes wxCopyFile hang, so do it the linux way
				mode_t oldUmask = umask(0);																				// Temporarily change umask so that permissions are copied exactly
				bool ans = mknod(To->GetFullPath().mb_str(wxConvUTF8), stat.GetStatstruct()->st_mode, stat.GetStatstruct()->st_rdev);	// Technically this works.  Goodness knows if it achieves anything useful!
				umask( oldUmask ); return ans==0;
			}

		return true;																													// A file has no children so we must have finished
	}
	
				// If From is a dir, we need to recreate a new subdir on .../.../To/, usually with the same name as the last segment of the 'From' one, but not if there was a name clash
				//	Notice we actually change To, since we want to put everything into the new subdir anyway, and this way that happens automatically
wxBusyCursor busy;				

if ( ! wxDirExists(From->GetFullPath()) )  return false;															// Shouldn't happen
															// Append the new pathname to the TO path. Use Assign not AppendDir, in case ToName is foo/bar/
wxString fullpath( To->GetPath() ); if ( fullpath.Last() != wxFILE_SEP_PATH ) fullpath << wxFILE_SEP_PATH;
if ( !ToName.IsEmpty() && ToName.Last() != wxFILE_SEP_PATH ) ToName << wxFILE_SEP_PATH;
To->Assign( fullpath + ToName );
if ( ! To->Mkdir( To->GetFullPath(), 0777, wxPATH_MKDIR_FULL ) )	 return false;			// Then use this to make a new dir
	
wxDir dir( From->GetPath() );																									// Use wxDir to deal sequentially with each child file & subdir
if ( !dir.IsOpened() )        return false;
if ( ! dir.HasSubDirs() && ! dir.HasFiles() )  return true;														// If the dir as no progeny, we're OK

wxString filename; bool AtLeastOneSuccess = false;
													// Go thru the dir, dealing first with files
bool cont = dir.GetFirst( &filename, wxEmptyString, wxDIR_FILES | wxDIR_HIDDEN );
while  ( cont )
	{	wxString file = wxFILE_SEP_PATH + filename;																// Make filename into /filename
	
		class FileData stat( From->GetPath() + file );																// Stat the file, to make sure it's not something else eg a symlink
		if ( stat.IsSymlink() )  																										// If it's a symlink, make a new one in destination dir
				{	wxString symlinktarget = stat.GetSymlinkDestination( true );				
					if ( RETAIN_REL_TARGET && symlinktarget.GetChar(0) != wxFILE_SEP_PATH )	// If it's a relative symlink (& we want to) we need to be clever to keep the same target
						{	wxFileName fn( symlinktarget ); fn.MakeAbsolute( From->GetPath() );		// So first make the target absolute, relative to the original symlink,
							fn.MakeRelativeTo( To->GetPath() );	symlinktarget = fn.GetFullPath();	//  then relative to the new one
						}				

					if ( CreateSymlink( symlinktarget, To->GetPath() + file ) )	AtLeastOneSuccess = true; 
				}
		 else
		 	if ( stat.IsRegularFile() )																								// If it's a normal file, use built-in wxCopyFile
				{	if ( wxCopyFile( From->GetPath() + file, To->GetPath() + file, false ) )		AtLeastOneSuccess = true; }
		 else
		 	{	mode_t oldUmask = umask(0);																				// Otherwise it's an oddity like a fifo, which makes wxCopyFile hang, so do it the linux way
				if ( ! mknod( wxString(To->GetPath() + file).mb_str(wxConvUTF8), stat.GetStatstruct()->st_mode, stat.GetStatstruct()->st_rdev ) )		AtLeastOneSuccess = true;	// Success returns 0
				umask( oldUmask );
			}
		cont = dir.GetNext(&filename);
	}
													// and now with any subdirs
cont = dir.GetFirst( &filename, wxEmptyString, wxDIR_DIRS | wxDIR_HIDDEN );													
while ( cont )
	{	wxString file = wxFILE_SEP_PATH + filename;																// Make filename into /filename
		FileData stat( From->GetPath() + file );																			// Stat the 'dir', to make sure it's not a symlink pointing to a dir!
		if ( stat.IsSymlink() )  
			{	if ( CreateSymlink( stat.GetSymlinkDestination(), To->GetPath() + file )	)  AtLeastOneSuccess = true; }	// If it's a symlink, make a new one in destination dir
 		 else
			{	wxFileName subdir( From->GetPath() + wxFILE_SEP_PATH + filename + wxFILE_SEP_PATH );	// Make the subdir into a wxFileName
				wxFileName copyofTo( To->GetPath() + wxFILE_SEP_PATH);									// Why? Because the wxFileName we pass will be altered in the subroutine, and To mustn't be
				if ( Paste( &subdir, &copyofTo, filename, true ) )		AtLeastOneSuccess = true;	// Do the subdir Paste by recursion
			}
		cont = dir.GetNext(&filename);
	}
	
return AtLeastOneSuccess;																									// which will be true if there has been >0 successes
}

void MyGenericDirCtrl::OnShortcutCopy(wxCommandEvent& WXUNUSED(event))				// Ctrl-C
{
size_t count;																			// No of selections (may be multiple, after all) 
wxArrayString paths;																// Array to store them

UnRedoManager::CloseSuperCluster();									// This switches off any extant supercluster, eg if last entry was a Cut.  Prevents inappropriate aggregation of next Paste

count = GetMultiplePaths( paths );											// Make the tree fill the array of selected pathname/filenames
if ( ! count ) return;																// No one at home

MyGenericDirCtrl::filearray = paths;										// Fill the "clipboard" with new data
MyGenericDirCtrl::FromID = GetId();										//  and the window ID of this pane
MyGenericDirCtrl::filecount = count;										// Store the count
MyGenericDirCtrl::Originpane = this;										// and which pane we were in: need if this is an archive

DoBriefLogStatus( count, wxEmptyString, _("copied") );
}

void MyGenericDirCtrl::OnShortcutPaste(wxCommandEvent& WXUNUSED(event))			// Ctrl-V
{
wxString destinationpath = GetPath();

									// First, check that we're not hovering over an empty area of a file-pane.  If so, we need to use its root path as the destination
if ( destinationpath.IsEmpty() && ( fileview == ISRIGHT ) )	// If we're in a fileview pane, but not hovering over something, 
			{	wxString rootpath = startdir;									//  get the root path for the pane
				if ( ! rootpath.IsEmpty() )
						destinationpath = rootpath;								// If it's valid, use it as the path
					else return;														// Otherwise, stop wasting time
			}
	
UnRedoManager::ContinueSuperCluster();							// Start a new cluster if from Copy, or continue the supercluster started by Cut
OnPaste( false, destinationpath );
UnRedoManager::EndCluster();
}

void MyGenericDirCtrl::OnPaste( bool FromDnD, wxString& destinationpath )					// Ctrl-V or DnD
{
enum DupRen WhattodoifClash = Unknown, WhattodoifCantRead = Unknown;		// These get passed to CheckDupRen.  May become SkipAll, OverwriteAll etc, but start with Unknown
size_t count;
wxArrayString filearray;
wxArrayInt IDs;
wxString path, DestPath;
MyGenericDirCtrl *originctrl, *destctrl;
												// We may have got here from Ctrl-V or Dnd, so get data into local vars as appropriate
if ( FromDnD )
	{	count = MyGenericDirCtrl::DnDfilecount;
		filearray = MyGenericDirCtrl::DnDfilearray;
		IDs.Add( MyGenericDirCtrl::DnDFromID ); IDs.Add( GetId() );	// Store the IDs of the panes that have changed
		path = MyGenericDirCtrl::DnDDestfilepath;
		originctrl = MyFrame::mainframe->GetActivePane(); destctrl = this;	// Active is the origin MyGenericDirCtrl, 'this' is the destination one
	}
 else
 	{	count = MyGenericDirCtrl::filecount;
		filearray = MyGenericDirCtrl::filearray;
		IDs.Add( MyGenericDirCtrl::FromID ); IDs.Add( GetId() );	// Store the IDs of the panes that have changed
		
		FileData fd( destinationpath );
		if ( fd.IsValid() && fd.IsSymlinktargetADir( true ) )			// If we're being asked to Paste onto a symlink, see if its target is a dir
			destinationpath = fd.GetUltimateDestination();			// If so, replace path with the symlink target.  NB If the target isn't a dir, we don't want to deref the symlink at all; just paste into its parent dir as usual

		path = destinationpath;													// OnShortcutPaste() processed the result of GetPath()
		originctrl = MyGenericDirCtrl::Originpane; destctrl = this;	// For Paste, the origin MyGenericDirCtrl was stored by Copy or Cut; 'this' is the destination one
	}

if ( ! count )  return; 																// Nothing in the "Clipboard"
if ( path.IsEmpty() )  return; 													// Nowhere to put it (not very likely, but nevertheless)

										// We need to check whether we're moving into or from within an archive: if so things are different
bool PastingFromTrash = filearray[0].StartsWith( DirectoryForDeletions::GetDeletedName() );	// We'll need this if we'd Cut from an archive ->trash; & now we're Pasting it
if  ( destctrl != NULL && destctrl->arcman != NULL && destctrl->arcman->IsArchive() )
	{	if ( arcman->GetArc()->MovePaste( originctrl, destctrl, filearray, path, false ) )	// If we're pasting into an archive (either from one or not) sort things out elsewhere
			DoBriefLogStatus( count, wxEmptyString, _("pasted") );
		return;
	}
if ( ! PastingFromTrash && originctrl != NULL && originctrl->arcman != NULL && originctrl->arcman->IsArchive() )	// Check we're not coming from trash, and originctrl is no longer relevant
	{	if ( originctrl->arcman->GetArc()->MovePaste( originctrl, destctrl, filearray, path, false ) )	 // If we're pasting FROM an archive into the outside world
			DoBriefLogStatus( filearray.GetCount(), wxEmptyString, _("pasted") );
		return;
	}

FileData fd( path );												// OK, back to non-archives
wxFileName Destination( path );												// Make sure the destination path is a dir with a terminal /  It's probably missing it
if ( (fd.IsDir() || fd.IsSymlinktargetADir(true))							// If the overall path is indeed a dir
 						&&  ! Destination.GetFullName().IsEmpty() )	
		path += Destination.GetPathSeparator();						//   plonk a / on the end of path, for the future
	else 
		{	Destination.SetFullName( wxEmptyString );				// If it IS a file, remove the filename, leaving the path for pasting to
			path = Destination.GetFullPath();
		}

FileData DestinationFD( path );
if ( ! DestinationFD.CanTHISUserWriteExec() )						// Make sure we have permission to write to the destination dir
	{	wxMessageDialog dialog(this, _("I'm afraid you don't have Permission to write to this Directory"), _("No Entry!"), wxOK | wxICON_ERROR );
		dialog.ShowModal(); return;
    }

FileData OriginFileFD( filearray[0] );
if ( ! OriginFileFD.CanTHISUserPotentiallyCopy() )				// Make sure we have exec permissions for the origin dir
	{	wxMessageDialog dialog(this, _("I'm afraid you don't have permission to access files from this Directory"),  _("No Exit!") ); return; }
		
int successes = 0;
for ( size_t c=0; c < count; c++ )											// For every path in the clipboard to be pasted
	{	wxString DestFilename;
		bool ItsADir;
															// We need to check we're not about to do something illegal or immoral
		CheckDupRen CheckIt( this, filearray[c], path );			// Make an instance of the class that tests for this
		CheckIt.IsMultiple = ( (c+1) < count );								// This flags whether there'll be more items after this one, so use different dialogs
		CheckIt.WhattodoifClash = WhattodoifClash; CheckIt.WhattodoifCantRead = WhattodoifCantRead;	// Set CheckIt's variables to match any previous choice
		
		if ( ! CheckIt.CheckForAccess() ) 									// This checks for Read permission failure, and sets a SkipAll flag if appropriate
			{ WhattodoifCantRead = CheckIt.WhattodoifCantRead; continue; }
		if ( CheckIt.CheckForIncest()  										// Check we're not pasting onto a descendant.  False means OK
				||  !CheckIt.CheckForPreExistence() )						//  or onto ourselves or a namesake (& if so offer to resolve). True means OK
			{	WhattodoifClash = CheckIt.WhattodoifClash;			// Store any new choice in local variable.  This catches SkipAll & Cancel, which return false
				if ( WhattodoifClash==Cancel ) { c = count; break; }	// Cancel this item & the rest of the loop by increasing c
				 	else continue; 													// Skip this item but continue with loop
			}
		 else 											// If there was a clash & subsequent Rename, CheckDupRen will have changed the filename or the name of the subdir to be made
		 	{	DestPath = CheckIt.finalpath;									// This will be either the original path without the filename/last dir-segment, or the altered version
		 		DestFilename = CheckIt.finalbit;								// Ditto, but the filename/last-segment
				WhattodoifClash = CheckIt.WhattodoifClash;			// Store any new choice in local variable
				ItsADir = CheckIt.ItsADir;										// CheckIt already tested whether it's a file or a dir.  Piggyback onto this, it saves constant rechecking
			}

		if ( ItsADir && filearray[c].Last() != wxFILE_SEP_PATH )
			filearray[c] += wxFILE_SEP_PATH;							// If we're pasting a dir, make sure it ends in a /
			
		wxFileName From( filearray[c] );										// Make the From path into a wxFileName
		wxFileName To( DestPath );											// and the To path
	
		if ( Paste( &From, &To, DestFilename, ItsADir ) )			// and paste it
			{												// NB The Paste has altered To, appending to it the pasted file or subdir. Fortunately, this is just what we want for Undoing
				wxString frompath = From.GetFullPath(), topath = To.GetFullPath();
				UnRedoPaste* UnRedoptr = new UnRedoPaste( frompath, IDs, topath, DestFilename, ItsADir );	// Make a new UnRedoPaste
				UnRedoManager::AddEntry( UnRedoptr );				// and store it for Undoing
				++successes;
			}
	}
if ( ! successes ) return;														// If none of the pastes worked, there's nothing left to do	

DoBriefLogStatus( successes, wxEmptyString, _("pasted") );
															// Finally we have to sort out the treectrl
SetPath( DestPath );																// We seem to need to tell the tree where the paste landed:  otherwise if it's in an unselected dir in the dirview, the selection won't alter & so the fileview won't be updated
MyFrame::mainframe->OnUpdateTrees( DestPath, IDs );		// Ask for the Refresh. DestPath is correct, whether a plain paste or a dup. (This wouldn't be true of path)
}

void MyGenericDirCtrl::OnShortcutUndo(wxCommandEvent& event)		// Ctrl-Z
{
UnRedoManager::UndoEntry();								// Do the Undoable
}

void MyGenericDirCtrl::OnShortcutRedo(wxCommandEvent& event)		// Ctrl-R
{
UnRedoManager::RedoEntry();								// ReDo the Undone
}

void DirGenericDirCtrl::OnToggleHidden(wxCommandEvent& event)		// Ctrl-H
{
bool hidden = ! GetShowHidden();														// Do the toggle
ShowHidden( hidden );																			//  & make it happen
partner->ShowHidden( hidden );															// Do it for fileview too, as it's difficult to imagine not wanting this

wxString path = GetPath();
UpdateStatusbarInfo( path );																// Make it appear in the statusbar
}

wxTreeItemId MyGenericDirCtrl::FindIdForPath( wxString& path )
{
wxTreeCtrl* tree = GetTreeCtrl();
wxTreeItemId root = GetRootId();

wxTreeItemId item = root;															// Since we don't know where to look in the tree, start from root
wxDirItemData *data = (wxDirItemData *) tree->GetItemData( item );
//wxLogTrace("Path="+path+"  data="+data->m_path);
while ( data->m_path != path )
	{	item = tree->GetNext( item );												// Get the next leaf in the tree
		if ( ! item.IsOk() ) return item;											// If item is corrupt, return as invalid
		if ( item==root ) return wxTreeItemId();							// If we've traversed the whole tree without success, signal by returning invalid
		data = (wxDirItemData *) tree->GetItemData( item );		// OK, try this leaf
/*wxLogTrace("data="+data->m_path);*/	}
//	if (item.IsOk() ) wxLogTrace("Found it");
return item;																						// If we're here, must have matched	
}

void DirGenericDirCtrl::ReloadTree( wxString path, wxArrayInt& IDs )	// Either F5 or after an Cut/Paste/Del/UnRedo, to keep treectrl congruent with reality
{
	// First decide whether it's important for this pane to be refreshed ie if it is/was the active pane, or otherwise if a file/dir that's just been acted upon is visible here
bool flag = false;
wxWindowID pane = GetId();																			// This is us
for ( unsigned int c=0; c < IDs.GetCount(); c++ )										// For every ID passed here,
	if ( IDs[c] == pane )	{ flag = true; break; }												//  see if it's us

		// There are 2 possibilities.  Either the dir that's just been acted upon is visible, or its not.  Remember, invisible means no TreeItemId exists
wxTreeItemId item = FindIdForPath( path );													// Translate path into wxTreeItemId
if ( ! item.IsOk() )																								// If it's not valid,
	{	if ( path  != wxFILE_SEP_PATH )															//    & it's not root!
			{	while ( path.Right(1) == wxFILE_SEP_PATH && path.Len() > 1 )	//       and if it's got a terminal '/' or two 
								path.RemoveLast();														// 		  remove it and try again without.
				item = FindIdForPath( path );															// This is cos the last dir of a branch doesn't have them, yet path DOES, so would never match
			}
	}

if ( item.IsOk() )																								// If it's valid, the item is visible, so definitely refresh it
	{	wxTreeCtrl* tree = GetTreeCtrl();
		bool subs = true;																					// If this is the root, we need a +/- box, otherwise there is only a dangling line coming from nowhere
		if ( tree->GetItemParent(item) != GetRootId() )
			{	if ( arcman != NULL && arcman->IsArchive() )
					{	FakeDir* fakedir = arcman->GetArc()->Getffs()->GetRootDir()->FindSubdirByName( path );
							subs = (fakedir != NULL && fakedir->SubDirCount() > 0);
					}
				  else	subs = HasThisDirSubdirs( path ); // It saves a lot of faffing with the treectrl if we use this global function to do a real-life count of subdirs --- remember this is a dirview so children must be dirs
			}
		tree->SetItemHasChildren( item, subs );				

		if ( ((wxDirItemData *) tree->GetItemData( item ))->m_isExpanded ) tree->Collapse(item);	// Do the refresh by (collapsing if needed &) expanding	
		tree->Expand(item);
	}

 else 																			// Otherwise the path isn't visible
 	if ( flag )																										// If the pane must be refreshed,
 		ExpandPath( path ); 																				// 	this should make it so

if ( path == GetPath() || (path+wxFILE_SEP_PATH) == GetPath() )				// (Note another '/' problem) Finally, if the altered item is one selected on this pane,
	partner->ReCreateTreeFromSelection();													// 	tell file-partner to refresh
}


BEGIN_EVENT_TABLE(DirGenericDirCtrl,MyGenericDirCtrl)
	EVT_MENU(SHCUT_NEW, DirGenericDirCtrl::NewFile)
	EVT_MENU(SHCUT_TOGGLEHIDDEN, DirGenericDirCtrl::OnToggleHidden)
	EVT_CONTEXT_MENU(DirGenericDirCtrl::ShowContextMenu)
END_EVENT_TABLE()

