/////////////////////////////////////////////////////////////////////////////
// (Original) Name:        dirctrlg.h
// Purpose:     wxGenericDirCtrl class
//              Builds on wxDirCtrl class written by Robert Roebling for the
//              wxFile application, modified by Harm van der Heijden.
//              Further modified for Windows. Further modified for 4Pane
// Author:      Robert Roebling, Harm van der Heijden, Julian Smart et al
// Modified by: David Hart 2003-7
// Created:     21/3/2000
// Copyright:   (c) Robert Roebling, Harm van der Heijden, Julian Smart. Alterations (c) David Hart
// Licence:     wxWindows licence
/////////////////////////////////////////////////////////////////////////////
/*
#ifndef _WX_DIRCTRL_H_
#define _WX_DIRCTRL_H_

#if defined(__GNUG__) && !defined(__APPLE__)
#pragma interface "dirctrlg.h"
#endif

#if wxUSE_DIRDLG
*/

#ifndef MYGENERICDIRCTRL
#define MYGENERICDIRCTRL


#include "wx/treectrl.h"
#include "wx/dialog.h"
#include <wx/dirctrl.h>
#include "wx/choice.h"

#include "Externs.h"							// //
#include "ArchiveStream.h"				// //

//-----------------------------------------------------------------------------
// classes
//-----------------------------------------------------------------------------

class WXDLLEXPORT wxTextCtrl;
class DirGenericDirCtrl;

//-----------------------------------------------------------------------------
// Extra styles for wxGenericDirCtrl
//-----------------------------------------------------------------------------

/*enum
{
    // Only allow directory viewing/selection, no files
    wxDIRCTRL_DIR_ONLY       = 0x0010,
    // When setting the default path, select the first file in the directory
    wxDIRCTRL_SELECT_FIRST   = 0x0020,
    // Show the filter list
    wxDIRCTRL_SHOW_FILTERS   = 0x0040,
    // Use 3D borders on internal controls
    wxDIRCTRL_3D_INTERNAL    = 0x0080,
    // Editable labels
    wxDIRCTRL_EDIT_LABELS    = 0x0100
};
*/
//-----------------------------------------------------------------------------
// wxDirItemData
//-----------------------------------------------------------------------------
/*
class WXDLLEXPORT wxDirItemData : public wxTreeItemData
{
public:
    wxDirItemData(const wxString& path, const wxString& name, bool isDir);
    ~wxDirItemData();
    void SetNewDirName(const wxString& path);

    bool HasSubDirs() const;
    bool HasFiles(const wxString& spec = wxEmptyString) const;

    wxString m_path, m_name;
    bool m_isHidden;
    bool m_isExpanded;
    bool m_isDir;
};
*/
//-----------------------------------------------------------------------------
// wxDirCtrl
//-----------------------------------------------------------------------------

class WXDLLEXPORT wxDirFilterListCtrl;

class DirSplitterWindow;	// //
class MyTab;					// //
class MyTreeCtrl;				// //

class MyGenericDirCtrl : public wxGenericDirCtrl
{
public:
    MyGenericDirCtrl();
    MyGenericDirCtrl(wxWindow *parent, MyGenericDirCtrl* parentcontrol, const wxWindowID id = -1,
              const wxString &dir = wxDirDialogDefaultFolderStr,
              const wxPoint& posit = wxDefaultPosition,
              const wxSize& sze = wxDefaultSize,
              long style = wxDIRCTRL_3D_INTERNAL|wxSUNKEN_BORDER,
              const wxString& filter = wxEmptyString,
              int defaultFilter = 0,
              const wxString& name = wxT("MyGenericDirCtrl"),
				bool file_view = ISLEFT, bool full_tree=false )	: pos( posit ), size( sze )
    {
		ParentControl = parentcontrol;	// //
        Init();
		 startdir = dir;	// //
		 fileview = file_view;	// //
		 fulltree = full_tree;		// //
        Create(parent, id, dir, style, filter, defaultFilter, name);
    }
    
    bool Create(wxWindow *parent, const wxWindowID id = -1,
       			     	const wxString &dir = wxDirDialogDefaultFolderStr,
         			   	long style = wxDIRCTRL_3D_INTERNAL|wxSUNKEN_BORDER,
              			const wxString& filter = wxEmptyString,
              			int defaultFilter = 0,
              			const wxString& name = wxTreeCtrlNameStr
						);
void FinishCreation();		// // Separated from Create() so that MyFileDirCtrl can be initialised first			
void CreateTree();				// //	I extracted this from Create, 2b reused in ReCreate
	
    virtual void Init();

    virtual ~MyGenericDirCtrl();

ArchiveStreamMan* arcman;											// //	Manages any archive stream
		
wxString  startdir;															// //	Holds either the (real) root or the selected dir, depending on fulltree
MyGenericDirCtrl* partner;												// //	Pointer to the twin pane, for coordination of dir selection
bool isright;																	// //	Are we part of the left/top twinpane, or the right/bottom?
void  TakeYourPartners( MyGenericDirCtrl *otherside )	// //	Set the pointer to this ctrl's partner
		{ partner = otherside; }
    void OnExpandItem( wxTreeEvent &event );
    void OnCollapseItem( wxTreeEvent &event );
	void OnBeginEditItem( wxTreeEvent &event );
	void OnEndEditItem( wxTreeEvent &event );
    void OnSize( wxSizeEvent &event );
void OnTreeSelected( wxTreeEvent &event );			// //
void OnTreeItemDClicked( wxTreeEvent &event );	// //
	void OnOutsideSelect( wxString& path, bool CheckDevices=true );		// // Used to redo Ctrl from outside, eg the floppy-button clicked (& if called from Mount, don't remount again!)
	void RefreshTree( wxString& path, bool CheckDevices=true );			// // Refreshes the display to show any outside changes (or following a change of rootitem via OnOutsideSelect)
	void OnRefreshTree(wxCommandEvent& event);								// // From Context menu or F5. Needed to provide the parameter for RefreshTree
void DisplaySelectionInTBTextCtrl( wxString& selection );	// // 
virtual void NewFile(wxCommandEvent& event){};				// // Create a new File or Dir

wxString GetActiveDirectory();									// // Returns the root dir if in branch-mode, or the selected dir in Fulltree-mode
wxString GetActiveDirPath();										// // Returns selected dir in  either mode  NB GetPath returns either dir OR file

    // Try to expand as much of the given path as possible.
    virtual bool ExpandPath(const wxString& path);

    // Accessors

    virtual inline wxString GetDefaultPath() const { return m_defaultPath; }
    virtual void SetDefaultPath(const wxString& path) { m_defaultPath = path; }

    // Get dir or filename
    virtual wxString GetPath() const;
virtual size_t GetMultiplePaths( wxArrayString& paths ) const; // // Get an array of selected paths, used as multiple selection is enabled    

static bool Clustering;					// // Flags that we're in the middle of a cluster of activity, so don't update trees yet

static wxArrayString filearray;		// // These form the "clipboard" for copy etc.  Static as we want only one for whole application,
static size_t filecount;
static wxWindowID FromID;
static MyGenericDirCtrl* Originpane;
static wxArrayString DnDfilearray;	// // Similarly for DragnDrop
static wxString DnDDestfilepath;	// // Where we're pasting to
static wxWindowID DnDFromID;	// // ID of the originating pane
static wxWindowID DnDToID;		// // ID of the destination pane
static size_t DnDfilecount;
	
    // Get selected filename path only (else empty string).
    // I.e. don't count a directory as a selection
    virtual wxString GetFilePath() const;
    virtual void SetPath(const wxString& path);
    
    virtual void ShowHidden( bool show );
    virtual bool GetShowHidden() { return m_showHidden; }
void SetShowHidden( bool show ){ m_showHidden = show; }
// //   virtual wxString GetFilter() const { return m_filter; }
// //   virtual void SetFilter(const wxString& filter);
void SetFilterArray( const wxArrayString& array, bool FilesOnly = false ){ FilterArray = array; DisplayFilesOnly = FilesOnly; }	// //
void SetFilterArray( const wxString& filterstring, bool FilesOnly = false );	// // Parse the string into individual filters & store in array
wxString GetFilterString();																			// //
wxArrayString& GetFilterArray(){ return FilterArray; }									// //
bool DisplayFilesOnly;																				// // Flags a fileview to filter out dirs if required
bool NoVisibleItems;																					// // Flags whether the user has filtered out all the items. Used for menuitem updateui
// //  virtual int GetFilterIndex() const { return m_currentFilter; }
// //   virtual void SetFilterIndex(int n);

    virtual wxTreeItemId GetRootId() { return m_rootId; }

    virtual wxTreeCtrl* GetTreeCtrl() const { return (wxTreeCtrl*)m_treeCtrl; }				// // Altered as m_treeCtrl is now a MyTreeCtrl
// //    virtual wxDirFilterListCtrl* GetFilterListCtrl() const { return m_filterListCtrl; }

    // Helper
    virtual void SetupSections();
    
    // Parse the filter into an array of filters and an array of descriptions
// //    virtual int ParseFilter(const wxString& filterStr, wxArrayString& filters, wxArrayString& descriptions);
    
    // Find the child that matches the first part of 'path'.
    // E.g. if a child path is "/usr" and 'path' is "/usr/include"
    // then the child for /usr is returned.
    // If the path string has been used (we're at the leaf), done is set to TRUE
    virtual wxTreeItemId FindChild(wxTreeItemId parentId, const wxString& path, bool& done);
    
    // Resize the components of the control
    virtual void DoResize();
    
    // Collapse & expand the tree, thus re-creating it from scratch:
    virtual void ReCreateTree();

bool fulltree;									// //	Flags whether we're allowed to show truncated trees, or if always full
bool SelFlag;								// //	Flags if an apparent selection event should be disabled as its coming from an internal process
bool fileview;								// //	Is this a left-side, directory, view or a right-side, file view?
	
void ReCreateTreeFromSelection();	// // On selection/DClick event, this redoes the tree & reselects
MyTab* ParentTab;							// // Pointer to the control's real parent, a tab panel.  Managed by derived class	

protected:
    void ExpandDir(wxTreeItemId parentId);
    void CollapseDir(wxTreeItemId parentId);
#if wxVERSION_NUMBER < 2500
    void AddSection(const wxString& path, const wxString& name, int imageId = 0);
#else
    const wxTreeItemId AddSection(const wxString& path, const wxString& name, int imageId = 0);
#endif
    //void FindChildFiles(wxTreeItemId id, int dirFlags, wxArrayString& filenames);

    // Extract description and actual filter from overall filter string
    bool ExtractWildcard(const wxString& filterStr, int n, wxString& filter, wxString& description);

void SetTreePrefs();			// //  Adjusts the Tree width parameters according to prefs

MyGenericDirCtrl* ParentControl;		// //	Pointer to the derived CONTROL, eg DirGenericDirCtrl, not the splitter window or tab panel
			// ----------------------------------------------The following were originally in MyDirs/Files	--------------------------------------------------------------------------------
wxTreeItemId FindIdForPath( wxString& path );
void OnShortcutUndo(wxCommandEvent&);														// // Ctrl-Z
void OnShortcutRedo(wxCommandEvent&);														// // Ctrl-R
void Delete( bool trash, bool fromCut=false );													// // Used by both OnShortcutTrash and OnShortcutDel, & by Cut

public:
void OnSplitpaneVertical(wxCommandEvent& WXUNUSED(event));					// // Vertically splits the panes of the current tab
void OnSplitpaneHorizontal(wxCommandEvent& WXUNUSED(event));				// // Horizontally splits the panes of the current tab
void OnSplitpaneUnsplit(wxCommandEvent& WXUNUSED(event));					// // Unsplits the panes of the current tab
void OnFilter(wxCommandEvent& WXUNUSED(event));									// // Calls Filter dialog
void OnProperties(wxCommandEvent& WXUNUSED(event));								// // Calls Properties dialog
void OnBrowseButton(wxCommandEvent& WXUNUSED(event));						// // If symlink-target Browse button clicked in Properties dialog
void OnLinkTargetButton(wxCommandEvent& event);											// // If symlink-target button clicked in Properties dialog
void LinkTargetButton(wxCommandEvent& event);												// // Recreate tree to goto selected link's target 
void BrowseButton();																							// // If symlink-target Browse button clicked in Properties dialog 
wxDialog* PropertyDlg;																						// // To store the above dlg

void OnShortcutGoToSymlinkTarget(wxCommandEvent& event){ GoToSymlinkTarget( event.GetId() == SHCUT_GOTO_ULTIMATE_SYMLINK_TARGET ); }
void GoToSymlinkTarget( bool ultimate );															// // Goto selected link's target (or ultimate target).  Called both by the above context-menu entry & by LinkTargetButton()

bool OnExitingArchive();																					// // We were displaying the innards of an archive, and now we're not

void OnShortcutTrash(wxCommandEvent& WXUNUSED(event));						// // Del
void OnShortcutDel(wxCommandEvent& WXUNUSED(event));							// // Sh-Del
void OnShortcutCopy(wxCommandEvent& WXUNUSED(event));						// // Ctrl-C
void OnShortcutCut(wxCommandEvent& WXUNUSED(event));							// // Ctrl-X
void OnShortcutHardLink(wxCommandEvent& WXUNUSED(event));					// // Ctr-Sh-L
void OnShortcutSoftLink(wxCommandEvent& WXUNUSED(event));					// // Alt-Sh-L
void OnShortcutPaste(wxCommandEvent& WXUNUSED(event));						// // Ctrl-V.  Calls the following:
void OnPaste( bool FromDnD, wxString& destinationpath = (wxString&)wxEmptyString );	// // Ctrl-V or DnD
void OnReplicate(wxCommandEvent& event);						// // Duplicates this pane's filepath in its opposite pane
void OnSwapPanes(wxCommandEvent& event);					// // Swap this pane's filepath with that of its opposite pane
void OnLink( bool FromDnD, int linktype );								// // Create hard or soft link
void OnDnDMove();																// //
MyGenericDirCtrl* GetOppositeTwinPane();							// // Return a pointer to the twin-pane opposite this one
void RefreshVisiblePanes(wxFont* font=(wxFont*)NULL);		// // Called e.g. after a change of font
void ReloadDirviewToolbars();												// // Called after a change of user-defined tools
static bool Paste(wxFileName* From, wxFileName* To, wxString& ToName, bool ItsADir );	// // Copies files, dirs+contents.  Static as used by UnRedo too
static bool ReallyDelete( wxFileName* PathName );															// // Not just trashing, the real thing.  eg for undoing pastes
static bool Move(wxFileName* From, wxFileName* To, wxString ToName);							// // Moves files, dirs+contents.  Static as used by UnRedo too
int OnNewItem( wxString& newname, wxString& path, bool ItsADir = true );						// // Used to create both new dirs & files
void OnDup(wxCommandEvent& WXUNUSED(event));														// // Duplicate.  Actually calls DoRename
void OnRename(wxCommandEvent& WXUNUSED(event));												// // F2 for rename. Calls DoRename
void DoRename( bool RenameOrDuplicate );								// // Does either rename or duplicate, depending on the bool
void DoMultipleRename( wxArrayString& paths, bool duplicate);	// // Renames (or duplicates) multiple files/dirs
int Rename( wxString& path, wxString& origname, wxString& newname, bool ItsADir, bool dup=false );	// // Renames both dirs & files (or dups if the bool is set)
			// -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
private:
    bool            m_showHidden;
    wxTreeItemId    m_rootId;
    wxImageList*    m_imageList;
    wxString        m_defaultPath; // Starting path
    long            m_styleEx; // Extended style
    wxString        m_filter;  // Wildcards in same format as per wxFileDialog	// // No longer used
    int             m_currentFilter; // The current filter index					// // No longer used
    wxString m_currentFilterStr; 		// Current filter string				// // No longer used, except as a convenient wxString
    wxArrayString FilterArray; 			// // Flter strings now they can be multiple
    MyTreeCtrl*     m_treeCtrl;
    wxDirFilterListCtrl* m_filterListCtrl;	// // No longer used

long treeStyle;		// //	I moved these 4 vars to here from locals in Create, so their data is still available in ReCreate 
long filterStyle;
const wxPoint& pos;
const wxSize& size;																								
	
//	wxTreeItemId LocateItem( wxDirItemData* data );				// //	MINE. Helper function for OnTreeSelected()
//	int CompareItems( wxTreeItemId candidateId, wxDirItemData* data );		// Subfunction for LocateItem()

private:
    DECLARE_EVENT_TABLE()
    DECLARE_DYNAMIC_CLASS(MyGenericDirCtrl)
};



//-----------------------------------------------------------------------------
// wxDirFilterListCtrl
//-----------------------------------------------------------------------------
/*
class WXDLLEXPORT wxDirFilterListCtrl: public wxChoice
{
public:
    wxDirFilterListCtrl() { Init(); }
    wxDirFilterListCtrl(wxGenericDirCtrl* parent, const wxWindowID id = -1,
              const wxPoint& pos = wxDefaultPosition,
              const wxSize& size = wxDefaultSize,
              long style = 0)
    {
        Init();
        Create(parent, id, pos, size, style);
    }
    
    bool Create(wxGenericDirCtrl* parent, const wxWindowID id = -1,
              const wxPoint& pos = wxDefaultPosition,
              const wxSize& size = wxDefaultSize,
              long style = 0);

    void Init();

    ~wxDirFilterListCtrl() {};

    //// Operations
    void FillFilterList(const wxString& filter, int defaultFilter);

    //// Events
    void OnSelFilter(wxCommandEvent& event);

protected:
    wxGenericDirCtrl*    m_dirCtrl;

    DECLARE_EVENT_TABLE()
    DECLARE_CLASS(wxDirFilterListCtrl)
};
*/
#if !defined(__WXMSW__) && !defined(__WXMAC__) && !defined(__WXPM__)
    #define wxDirCtrl wxGenericDirCtrl
#endif

// Symbols for accessing individual controls
#define wxID_TREECTRL          7000
#define wxID_FILTERLISTCTRL    7001

//#endif // wxUSE_DIRDLG

#endif
    // MYGENERICDIRCTRL  (Originally  _WX_DIRCTRLG_H_)
