/////////////////////////////////////////////////////////////////////////////
// Name:        MyTreeCtrl.cpp
// Purpose:    Tree and Treelistctrl.
// Derived:     The listctrl parts adapted from wxTreeListCtrl (c) Alberto Griggio, Otto Wyss
// See:            http://wxcode.sourceforge.net/components/treelistctrl/
// Part of:       4Pane
// Author:       David Hart
// Copyright:   (c) 2007 David Hart
// Licence:     wxWindows licence
/////////////////////////////////////////////////////////////////////////////

#ifndef WX_PRECOMP
    #include "wx/log.h"
	#include <wx/dirctrl.h>
	#include "wx/dragimag.h"
#endif

#if wxVERSION_NUMBER > 2500
	#include "wx/renderer.h"
#endif


#include "wx/filename.h" 


#ifdef __WXGTK__
    #include <gtk/gtk.h>
	#if wxVERSION_NUMBER < 2701
    	#include <wx/gtk/win_gtk.h>
	#endif
#endif

#include "MyGenericDirCtrl.h"
#include "Dup.h"
#include "MyFiles.h"
#include "MyDirs.h"
#include "MyFrame.h"
#include "Filetypes.h"
#include "Redo.h"
#include "Devices.h"
#include "MyTreeCtrl.h"
#include "Accelerators.h"

#include "bitmaps/include/columnheaderdown.xpm"	// // Icons for the fileview columns, selected & selected-reverse-sort 
#include "bitmaps/include/columnheaderup.xpm"

#if wxVERSION_NUMBER < 2900	// to avoid compiler warnings
	#define wxUSE_NEW_DC 0
#endif

class WXDLLEXPORT wxGenericTreeItem;

WX_DEFINE_EXPORTED_ARRAY(wxGenericTreeItem *, wxArrayGenericTreeItems);

enum { ColDownIcon = 0, ColUpIcon };					// These are the indices of the Down/Up column icons within HeaderimageList

class WXDLLEXPORT wxGenericTreeItem
{
public:
    // ctors & dtor
    wxGenericTreeItem() { m_data = NULL; }
    wxGenericTreeItem( wxGenericTreeItem *parent,
                       const wxString& text,
                       int image,
                       int selImage,
                       wxTreeItemData *data );

    ~wxGenericTreeItem();

    // trivial accessors
    wxArrayGenericTreeItems& GetChildren() { return m_children; }

    const wxString& GetText() const { return m_text; }
    int GetImage(wxTreeItemIcon which = wxTreeItemIcon_Normal) const
        { return m_images[which]; }
    wxTreeItemData *GetData() const { return m_data; }

    // returns the current image for the item (depending on its
    // selected/expanded/whatever state)
    int GetCurrentImage() const;

    void SetText( const wxString &text );
    void SetImage(int image, wxTreeItemIcon which) { m_images[which] = image; }
    void SetData(wxTreeItemData *data) { m_data = data; }

    void SetHasPlus(bool has = true) { m_hasPlus = has; }

    void SetBold(bool bold) { m_isBold = bold; }

    int GetX() const { return m_x; }
    int GetY() const { return m_y; }

    void SetX(int x) { m_x = x; }
    void SetY(int y) { m_y = y; }

    int  GetHeight() const { return m_height; }
    int  GetWidth()  const { return m_width; }

    void SetHeight(int h) { m_height = h; }
    void SetWidth(int w) { m_width = w; }

    wxGenericTreeItem *GetParent() const { return m_parent; }

    // operations
        // deletes all children notifying the treectrl about it if !NULL
        // pointer given
    void DeleteChildren(wxGenericTreeCtrl *tree = NULL);

    // get count of all children (and grand children if 'recursively')
    size_t GetChildrenCount(bool recursively = true) const;

    void Insert(wxGenericTreeItem *child, size_t index)
    { m_children.Insert(child, index); }

    void GetSize( int &x, int &y, const wxGenericTreeCtrl* );

        // return the item at given position (or NULL if no item), onButton is
        // true if the point belongs to the item's button, otherwise it lies
        // on the button's label
    wxGenericTreeItem *HitTest( const wxPoint& point,
                                const wxGenericTreeCtrl *,
                                int &flags,
                                int level );

    void Expand() { m_isCollapsed = FALSE; }
    void Collapse() { m_isCollapsed = true; }

    void SetHilight( bool set = true ) { m_hasHilight = set; }

    // status inquiries
    bool HasChildren() const { return !m_children.IsEmpty(); }
    bool IsSelected()  const { return m_hasHilight != 0; }
    bool IsExpanded()  const { return !m_isCollapsed; }
    bool HasPlus()     const { return m_hasPlus || HasChildren(); }
    bool IsBold()      const { return m_isBold != 0; }

    // attributes
        // get them - may be NULL
    wxTreeItemAttr *GetAttributes() const { return m_attr; }
        // get them ensuring that the pointer is not NULL
    wxTreeItemAttr& Attr()
    {
        if ( !m_attr )
        {
            m_attr = new wxTreeItemAttr;
            m_ownsAttr = true;
        }
        return *m_attr;
    }
        // set them
    void SetAttributes(wxTreeItemAttr *attr)
    {
        if ( m_ownsAttr ) delete m_attr;
        m_attr = attr;
        m_ownsAttr = FALSE;
    }
        // set them and delete when done
    void AssignAttributes(wxTreeItemAttr *attr)
    {
        SetAttributes(attr);
        m_ownsAttr = true;
    }

private:
    // since there can be very many of these, we save size by chosing
    // the smallest representation for the elements and by ordering
    // the members to avoid padding.
    wxString            m_text;         // label to be rendered for item

    wxTreeItemData     *m_data;         // user-provided data
#if wxVERSION_NUMBER >= 2900
    int                 m_state;        // item state New in 2.9, and causes a crash if it's not there. Presumably because the object size is changed
    int                 m_widthText;
    int                 m_heightText;
#endif
    wxArrayGenericTreeItems m_children; // list of children
    wxGenericTreeItem  *m_parent;       // parent of this item

    wxTreeItemAttr     *m_attr;         // attributes???

    // tree ctrl images for the normal, selected, expanded and
    // expanded+selected states
#if wxVERSION_NUMBER < 2504
    short               m_images[wxTreeItemIcon_Max];
#else
    int                 m_images[wxTreeItemIcon_Max];
#endif
    wxCoord             m_x;            // (virtual) offset from top
    wxCoord             m_y;            // (virtual) offset from left
#if wxVERSION_NUMBER < 2504
    short               m_width;        // width of this item
    unsigned char       m_height;       // height of this item
#else
    int                 m_width;        // width of this item
    int                 m_height;       // height of this item
#endif
    // use bitfields to save size
    int                 m_isCollapsed :1;
    int                 m_hasHilight  :1; // same as focused
    int                 m_hasPlus     :1; // used for item which doesn't have
                                          // children but has a [+] button
    int                 m_isBold      :1; // render the label in bold font
    int                 m_ownsAttr    :1; // delete attribute when done
};


//-----------------------------------------------------------------------------
//  Originally wxTreeListHeaderWindow  from wxTreeListCtrl 
//-----------------------------------------------------------------------------

#include <wx/arrimpl.cpp>
WX_DEFINE_OBJARRAY(wxArrayTreeListColumnInfo);

IMPLEMENT_DYNAMIC_CLASS(TreeListHeaderWindow,wxWindow);

static const int MINCOLWIDTH = 10;						// We don't want any displayed column to become so small as to be unresizable
static const int MINMAINCOLWIDTH = 100;				// We want the main column to be larger than the others

TreeListHeaderWindow::TreeListHeaderWindow()
{
    Init();

    m_owner = (MyTreeCtrl*) NULL;
    m_resizeCursor = (wxCursor*) NULL;
}

TreeListHeaderWindow::TreeListHeaderWindow( wxWindow *win,
						wxWindowID id,
						MyTreeCtrl* owner,
						const wxPoint& pos,
						const wxSize& size,
						long style,
						const wxString &name )
    : wxWindow( win, id, pos, size, style, name )
{
    Init();

    m_owner = owner;
    m_resizeCursor = new wxCursor(wxCURSOR_SIZEWE);

    SetBackgroundColour(wxSystemSettings::GetColour(	// //
			    wxSYS_COLOUR_BTNFACE));
}

TreeListHeaderWindow::~TreeListHeaderWindow()
{
    delete m_resizeCursor;
}

void TreeListHeaderWindow::Init()
{
    m_currentCursor = (wxCursor*) NULL;
    m_isDragging = FALSE;
    m_dirty = FALSE;
		
m_total_col_width = 0;										// // (Originally this wasn't given a default)
selectedcolumn = filename;									// // The next 2 should be set by config but let's give them sensible defaults anyway
reversesort = false;
}

#ifdef __WXGTK__
  #include "wx/gtk/dc.h"
#endif

void TreeListHeaderWindow::DoDrawRect( wxDC *dc, int x, int y, int w, int h )
{
#ifdef __WXGTK__
    GtkStateType state = m_parent->IsEnabled() ? GTK_STATE_NORMAL
                                               : GTK_STATE_INSENSITIVE;

    x = dc->LogicalToDeviceX( x );
#if wxVERSION_NUMBER >= 2701
// // The old-style way of doing things stopped working for gtk1.2 in 2.7.1 and above
// // as GTK_PIZZA(m_wxwindow)->bin_window returns rubbish, leading to a segfault
// // This was taken from wxRendererGTK::DrawHeaderButton in renderer.cpp, and works for both gtk.s
// // The wxUSE_NEW_DC is new in 2.9
	GdkWindow* gdk_window = NULL;
#if wxUSE_NEW_DC
    wxDCImpl *impl = dc->GetImpl();
    wxGTKDCImpl *gtk_impl = wxDynamicCast( impl, wxGTKDCImpl );
    if (gtk_impl)
         gdk_window = gtk_impl->GetGDKWindow();
#else
     gdk_window = dc->GetGDKWindow();
#endif

    wxASSERT_MSG( gdk_window,
                  wxT("cannot use wxRendererNative on wxDC of this type") );
    gtk_paint_box (m_wxwindow->style, gdk_window,
                   state, GTK_SHADOW_OUT,
                   (GdkRectangle*) NULL, m_wxwindow, "button",
                   x-1, y-1, w+2, h+2);
#else
    gtk_paint_box (m_wxwindow->style, GTK_PIZZA(m_wxwindow)->bin_window,
                   state, GTK_SHADOW_OUT,
                   (GdkRectangle*) NULL, m_wxwindow, "button",
                   x-1, y-1, w+2, h+2);
#endif // wxVERSION_NUMBER >= 2701

#elif defined( __WXMAC__  )
    const int m_corner = 1;

    dc->SetBrush( *wxTRANSPARENT_BRUSH );

    dc->SetPen( wxPen(wxSystemSettings::GetSystemColour(
			  wxSYS_COLOUR_BTNSHADOW), 1, wxSOLID));
    dc->DrawLine( x+w-m_corner+1, y, x+w, y+h );  // right (outer)
    dc->DrawRectangle( x, y+h, w+1, 1 );          // bottom (outer)

    wxPen pen( wxColour( 0x88 , 0x88 , 0x88 ), 1, wxSOLID );

    dc->SetPen( pen );
    dc->DrawLine( x+w-m_corner, y, x+w-1, y+h );  // right (inner)
    dc->DrawRectangle( x+1, y+h-1, w-2, 1 );      // bottom (inner)

    dc->SetPen( *wxWHITE_PEN );
    dc->DrawRectangle( x, y, w-m_corner+1, 1 );   // top (outer)
    dc->DrawRectangle( x, y, 1, h );              // left (outer)
    dc->DrawLine( x, y+h-1, x+1, y+h-1 );
    dc->DrawLine( x+w-1, y, x+w-1, y+1 );
#else // !GTK, !Mac
    const int m_corner = 1;

    dc->SetBrush( *wxTRANSPARENT_BRUSH );

    dc->SetPen( *wxBLACK_PEN );
    dc->DrawLine( x+w-m_corner+1, y, x+w, y+h );  // right (outer)
    dc->DrawRectangle( x, y+h, w+1, 1 );          // bottom (outer)

    wxPen pen(wxSystemSettings::GetColour(
		  wxSYS_COLOUR_BTNSHADOW ), 1, wxSOLID);

    dc->SetPen( pen );
    dc->DrawLine( x+w-m_corner, y, x+w-1, y+h );  // right (inner)
    dc->DrawRectangle( x+1, y+h-1, w-2, 1 );      // bottom (inner)

    dc->SetPen( *wxWHITE_PEN );
    dc->DrawRectangle( x, y, w-m_corner+1, 1 );   // top (outer)
    dc->DrawRectangle( x, y, 1, h );              // left (outer)
    dc->DrawLine( x, y+h-1, x+1, y+h-1 );
    dc->DrawLine( x+w-1, y, x+w-1, y+1 );
#endif
}

// shift the DC origin to match the position of the main window horz
// scrollbar: this allows us to always use logical coords
void TreeListHeaderWindow::AdjustDC(wxDC& dc)
{
    int xpix;
    m_owner->GetScrollPixelsPerUnit( &xpix, NULL );

    int x;
    m_owner->GetViewStart( &x, NULL );

    // account for the horz scrollbar offset
    dc.SetDeviceOrigin( -x * xpix, 0 );
}

void TreeListHeaderWindow::OnPaint( wxPaintEvent &WXUNUSED(event) )
{
    static const int HEADER_OFFSET_X = 1, HEADER_OFFSET_Y = 1;
#ifdef __WXGTK__
    wxClientDC dc( this );
#else
    wxPaintDC dc( this );
#endif

    PrepareDC( dc );
    AdjustDC( dc );

    // //dc.BeginDrawing();

    dc.SetFont( GetFont() );

    // width and height of the entire header window
    int w, h;
    GetClientSize( &w, &h );
    m_owner->CalcUnscrolledPosition(w, 0, &w, NULL);

    dc.SetBackgroundMode(wxTRANSPARENT);

    // do *not* use the listctrl colour for headers - one day we will have a
    // function to set it separately
    //dc.SetTextForeground( *wxBLACK );
    dc.SetTextForeground(wxSystemSettings::
                            GetColour( wxSYS_COLOUR_WINDOWTEXT ));			// //

    int x = HEADER_OFFSET_X;

    int numColumns = GetColumnCount();
    for ( int i = 0; i < numColumns && x < w; i++ )
    {
	wxTreeListColumnInfo& column = GetColumn(i);
        int wCol = column.GetWidth(); 

        // the width of the rect to draw: make it smaller to fit entirely
        // inside the column rect
        int cw = wCol - 2;

        dc.SetPen( *wxWHITE_PEN );

        DoDrawRect( &dc, x, HEADER_OFFSET_Y, cw, h-2 );

        // if we have an image, draw it on the right of the label	// // except that I want it in the middle --- see below 
        int image = column.GetImage(); //item.m_image;
	int ix = -2, iy = 0;
	wxImageList* imageList = m_owner->HeaderimageList;		// // m_owner->GetImageList();		I've had to make my own list, as my bitmaps are 16*10, not 16*16
        if(image != -1) {
            if(imageList) {
                imageList->GetSize(image, ix, iy);
	    }
            //else: ignore the column image
        }

        // extra margins around the text label
	static const int EXTRA_WIDTH = 3;
	static const int EXTRA_HEIGHT = 4;

	int text_width = 0;
	int text_x = x;
	// // int image_offset = cw - ix - 1;
	int image_offset = (cw - ix)/2;												// // The way to get a central image
	
	switch(column.GetAlignment()) {
	case wxTL_ALIGN_LEFT:
	    text_x += EXTRA_WIDTH;
	    cw -= ix + 2;
	    break;
	case wxTL_ALIGN_RIGHT:
	    dc.GetTextExtent(column.GetText(), &text_width, NULL);
	    text_x += cw - text_width - EXTRA_WIDTH;
	    image_offset = 0;
	    break;
	case wxTL_ALIGN_CENTER:
	    dc.GetTextExtent(column.GetText(), &text_width, NULL);
	    text_x += (cw - text_width)/2 + ix + 2;
	    image_offset = (cw - text_width - ix - 2)/2;
	    break;
	}

	// draw the image
	if(image != -1 && imageList) {
	    imageList->Draw(image, dc, x + image_offset/*cw - ix - 1*/,
			    HEADER_OFFSET_Y + (h - 4 - iy)/2,
			    wxIMAGELIST_DRAW_TRANSPARENT);
	}
	
        // draw the text clipping it so that it doesn't overwrite the column
        // boundary
        wxDCClipper clipper(dc, x, HEADER_OFFSET_Y, cw, h - 4 );

	dc.DrawText( column.GetText(),
                     text_x, HEADER_OFFSET_Y + EXTRA_HEIGHT );

        x += wCol;
    }

    // //dc.EndDrawing();
}

void TreeListHeaderWindow::DrawCurrent()
{
    int x1 = m_currentX;
    int y1 = 0;
    ClientToScreen( &x1, &y1 );

    int x2 = m_currentX-1;
#ifdef __WXMSW__
    ++x2; // but why ? 
#endif
    int y2 = 0;
    m_owner->GetClientSize( NULL, &y2 );
    m_owner->ClientToScreen( &x2, &y2 );

    wxScreenDC dc;
    dc.SetLogicalFunction( wxINVERT );
    dc.SetPen( wxPen( *wxBLACK, 2, wxSOLID ) );
    dc.SetBrush( *wxTRANSPARENT_BRUSH );

    AdjustDC(dc);

    dc.DrawLine( x1, y1, x2, y2 );

    dc.SetLogicalFunction( wxCOPY );

    dc.SetPen( wxNullPen );
    dc.SetBrush( wxNullBrush );
}

void TreeListHeaderWindow::OnMouse( wxMouseEvent &event )
{
    // we want to work with logical coords
    int x;
    m_owner->CalcUnscrolledPosition(event.GetX(), 0, &x, NULL);
    int y = event.GetY();

    if (m_isDragging)
    {
        SendListEvent(wxEVT_COMMAND_LIST_COL_DRAGGING,
                      event.GetPosition());

	// we don't draw the line beyond our window, but we allow dragging it
        // there
        int w = 0;
        GetClientSize( &w, NULL );
        m_owner->CalcUnscrolledPosition(w, 0, &w, NULL);
        w -= 6;

        // erase the line if it was drawn
        if ( m_currentX < w )
            DrawCurrent();

        if (event.ButtonUp())
        {
            ReleaseMouse();
            m_isDragging = FALSE;
            m_dirty = true;
						
size_t newwidth = wxMax( (m_currentX - m_minX), 10);							// // Find the new width --- may be negative, so don't let it get too small
int widthchange = newwidth - GetColumnWidth( m_column );
if ( m_column == GetLastVisibleCol() && newwidth-current > 0 )				// // If we're trying to drag-bigger the last visible column,
	{	newwidth += widthchange * 2;															// //    double the increment as the tiny drag-onto-able space makes it so slow otherwise
		m_columns[m_column].SetDesiredWidth( newwidth );						// //        and make it stick
	}

SetColumnWidth( m_column, newwidth );

	SizeColumns();	// //
            Refresh();
            SendListEvent(wxEVT_COMMAND_LIST_COL_END_DRAG,
                          event.GetPosition());
        }
        else
        {
            if (x > m_minX + 7)
                m_currentX = x;
            else
                m_currentX = m_minX + 7;

            // draw in the new location
            if ( m_currentX < w )
                DrawCurrent();
        }
    }
    else // not dragging
    {
#if wxVERSION_NUMBER > 2503														// // As ContextMenu events don't seem to get to the headerwindow otherwise
	if (event.RightUp())
  		{	wxContextMenuEvent conevent(wxEVT_CONTEXT_MENU, 0, ClientToScreen( event.GetPosition() ) );
			return ShowContextMenu( conevent );
		}
#endif
		m_minX = 0;
        bool hit_border = FALSE;

        // end of the current column
        int xpos = 0;

        // find the column where this event occured
        int countCol = GetColumnCount();
        for (int col = 0; col < countCol; col++)
        {
if ( IsHidden(col) )	 continue;															// // I've 'hidden' some cols by giving them widths of zero.  This takes it into account
            xpos += GetColumnWidth( col );
            m_column = col;

            if ( (abs(x-xpos) < 3) && (y < (HEADER_HEIGHT-1)) )			// // I've changed "y <" comparator to HEADER_HEIGHT-1: originally the figure 22 was given
            {
                // near the column border
                hit_border = true;
                break;
            }

            if ( x < xpos )
            {
                // inside the column
                break;
            }

            m_minX = xpos;
        }

        if (event.LeftDown() /*|| event.RightUp()*/)									// // I've grabbed RtUp earlier to fire the context menu.  What it was doing here anyway?
        {
            if (hit_border/* && event.LeftDown()*/)									// // See above comment
            {
                m_isDragging = true;
                m_currentX = x;
                DrawCurrent();
                CaptureMouse();
                SendListEvent(wxEVT_COMMAND_LIST_COL_BEGIN_DRAG,
                              event.GetPosition());
            }
            else // click on a column
            {
                SendListEvent( event.LeftDown()
			       ? wxEVT_COMMAND_LIST_COL_CLICK
			       : wxEVT_COMMAND_LIST_COL_RIGHT_CLICK,
			       event.GetPosition());
            }
        }
        else if (event.Moving())
        {
            bool setCursor;
            if (hit_border)
            {
                setCursor = m_currentCursor == wxSTANDARD_CURSOR;
                m_currentCursor = m_resizeCursor;
            }
            else
            {
                setCursor = m_currentCursor != wxSTANDARD_CURSOR;
                m_currentCursor = (wxCursor*)wxSTANDARD_CURSOR;
            }

            if ( setCursor )
                SetCursor(*m_currentCursor);
        }
    }
}

int TreeListHeaderWindow::GetNextVisibleCol( int currentcolumn )		// // Finds the next column with non-zero width
{
int colcount = GetColumnCount(); if ( !colcount || currentcolumn==(colcount-1) ) return -1;
for ( int col=currentcolumn+1; col < colcount; ++col )
	if ( ! IsHidden( col ) ) return col;
	
return -1;
}

int TreeListHeaderWindow::GetLastVisibleCol()		// // Finds the last column with non-zero width.  0 flags failure (but shouldn't happen!)
{
int colcount = GetColumnCount(); if ( !colcount ) return 0;	// Shouldn't happen as col 0 will always exist & be visible
for ( int col=colcount; col > 0; --col )
	if ( ! IsHidden( col-1 ) ) return col-1;
	
return 0;
}

int TreeListHeaderWindow::GetTotalPreferredWidth()	// // Omits the extra amount added to the last column to fill the whole client-size
{
int lastvisiblecol = GetLastVisibleCol();// if ( ! lastvisiblecol ) return GetWidth();
return GetWidth() - (GetColumnWidth( lastvisiblecol ) - GetColumnDesiredWidth( lastvisiblecol ));
}

void TreeListHeaderWindow::OnSetFocus( wxFocusEvent &WXUNUSED(event) )
{
    m_owner->SetFocus();
}

void TreeListHeaderWindow::SendListEvent(wxEventType type, wxPoint pos,
																													int column/*=-1*/ )	// // I've added int column, as otherwise we have a problem when the event refers to hiding a col
{
    wxWindow *parent = GetParent();
    wxListEvent le( type, parent->GetId() );
    le.SetEventObject( parent );
    le.m_pointDrag = pos;

    // the position should be relative to the parent window, not
    // this one for compatibility with MSW and common sense: the
    // user code doesn't know anything at all about this header
    // window, so why should it get positions relative to it?
    le.m_pointDrag.y -= GetSize().y;

    if ( column == -1 )																						// //
		le.m_col = m_column;
	 else le.m_col = -1;																						// // Otherwise pass an intentionally invalid column.  This tells FileGenericDirCtrl::HeaderWindowClicked() not to (re)reverse the column
	 
    parent->GetEventHandler()->ProcessEvent( le );
}


void TreeListHeaderWindow::AddColumn(const wxTreeListColumnInfo& col)
{
    m_columns.Add(col);
    m_total_col_width += col.GetWidth();
    //m_owner->GetHeaderWindow()->Refresh();
    //m_dirty = true;
//SizeColumns();																			// //
    m_owner->AdjustMyScrollbars();
    m_owner->m_dirty = true;
    Refresh();
}

void TreeListHeaderWindow::SetColumnWidth(size_t column, size_t width)
{
    if(column < GetColumnCount()) {
	m_total_col_width -= m_columns[column].GetWidth();
	m_columns[column].SetWidth(width);
	m_total_col_width += width;
	// // m_owner->AdjustMyScrollbars();  (Don't bother:  this is always done afterwards by the calling functions)
	m_owner->m_dirty = true;
	//m_dirty = true;
	Refresh();
    }
}

void TreeListHeaderWindow::InsertColumn(size_t before,
					  const wxTreeListColumnInfo& col)
{
    wxCHECK_RET(before < GetColumnCount(), _("Invalid column index"));
    m_columns.Insert(col, before);
    m_total_col_width += col.GetWidth();
    //m_dirty = true;
    //m_owner->GetHeaderWindow()->Refresh();
SizeColumns();																			// //
    // //m_owner->AdjustMyScrollbars();
    m_owner->m_dirty = true;
    Refresh();
}


void TreeListHeaderWindow::RemoveColumn(size_t column)
{
    wxCHECK_RET(column < GetColumnCount(), _("Invalid column"));
    m_total_col_width -= m_columns[column].GetWidth();
    m_columns.RemoveAt(column);
    //m_dirty = true;
SizeColumns();																			// //
    // //m_owner->AdjustMyScrollbars();
    m_owner->m_dirty = true;
    Refresh();
}


void TreeListHeaderWindow::SetColumn(size_t column,
				       const wxTreeListColumnInfo& info)
{
    wxCHECK_RET(column < GetColumnCount(), _("Invalid column"));
    size_t w = m_columns[column].GetWidth();
    m_columns[column] = info;
    //m_owner->GetHeaderWindow()->Refresh();
    //m_dirty = true;
    if(w != info.GetWidth()) {
	m_total_col_width += info.GetWidth() - w;
	SizeColumns();																			// //
	// //m_owner->AdjustMyScrollbars();
	m_owner->m_dirty = true;
    }
    Refresh();
}

void TreeListHeaderWindow::ToggleColumn( int col )		// // If it's hidden, show it & vice versa
{
if ( IsHidden( col ) )		ShowColumn( col );
 else ShowColumn( col, false );
}

void TreeListHeaderWindow::ShowColumn( int col, bool show /*=true*/ )		// // Show (or hide) the relevant column
{
if ( IsHidden( col ) == !show )  return;									// If it's already doing what we want, wave goodbye

int width;
if ( show )
	{	int last = GetLastVisibleCol();
		if ( col > last )																	// If the newly-shown column will become the last visible, 
			{	if ( GetColumnDesiredWidth( last ) > 0 )					//   we must adjust the current last col first, otherwise it'll never revert to its preferred width
					SetColumnWidth( last, GetColumnDesiredWidth( last ) );
				 else
					SetColumnWidth( last, m_owner->GetCharWidth() * DEFAULT_COLUMN_WIDTHS[last] );
			}
		width = GetColumnDesiredWidth( col );							// We need to know how big to make it
		if ( width <= 0 ) width = m_owner->GetCharWidth() * DEFAULT_COLUMN_WIDTHS[col];	// If there's no valid size, use default, adjusted for fontsize
	}
 else 
 	{	width = 0;																		// Hiding means zero width
		if ( GetSelectedColumn() == col )  
			{	SelectColumn(0);													// If this col is currently selected, select col 0 instead
				SendListEvent( wxEVT_COMMAND_LIST_COL_CLICK,  wxPoint(0,0), -2 );	// This tells the fileview not to re-order the files by col, as we just did
			}
 	}
	
SetColumnWidth( col, width );													// This also sets/resets the ishidden flag
if ( show ) SizeColumns();
  else ((FileGenericDirCtrl*)m_owner->GetParent())->DoSize();
m_dirty = true; 																		// We'll need to refresh the column display, as some names may be differently clipped
}

void TreeListHeaderWindow::SetSelectedColumn(size_t col)	// // Sets the selected column, the one that controls Sorting
{
GetColumn( GetSelectedColumn() ).SetImage( -1 );					// Turn off the image from the previously-selected column
selectedcolumn = (enum columntype)col;
}

void TreeListHeaderWindow::SelectColumn( int col )
{
ShowColumn( col );																// Better make sure it's visible!

SetSortOrder(0);																	// In case we were previously reverse-sorting, set to normal
SetSelectedColumn( col );
GetColumn(col).SetImage( ColDownIcon );
Refresh();
}

bool TreeListHeaderWindow::SetSortOrder( int order /*=-1*/ ) 	// // Set whether we're reverse-sorting.  If no parameter, negates sort-order
{
switch(order)
	{	case 0:	reversesort = false;	return reversesort; 			// If 0, set order to normal
	 	case 1:	reversesort = true;	return reversesort; 				// If 1, set order to reverse
		 default:	reversesort = ! reversesort; return reversesort; // Otherwise reverse the current order
	}
}

void TreeListHeaderWindow::ReverseSelectedColumn()
{
GetColumn( GetSelectedColumn() ).SetImage( ColDownIcon + SetSortOrder() );
Refresh();
}

void TreeListHeaderWindow::SizeColumns()		// // Resize the last visible col to occupy any spare space
{
size_t colcount = GetColumnCount(); if ( ! colcount ) return;

int width, h; GetClientSize(&width, &h);
int diff = width - m_total_col_width;										// Is the new width > or < than the previous value?
if ( diff == 0 ) return;																// If ISQ, not much to do

int lastvisiblecol = GetLastVisibleCol();
if ( diff < 0 ) 																			// If negative, reduce the last col's width towards its preferred value if currently higher
	{	int spare =  GetColumnWidth( lastvisiblecol ) - GetColumnDesiredWidth( lastvisiblecol );
		if ( spare > 0 )
			SetColumnWidth( lastvisiblecol, GetColumnWidth( lastvisiblecol ) - wxMin(abs(diff), spare) );
	}
 else
	SetColumnWidth( lastvisiblecol, GetColumnWidth( lastvisiblecol ) + diff );	// Otherwise just dump the spare space onto the last visible col, to avoid a silly-looking blank

m_owner->AdjustMyScrollbars();
Refresh(); 
}

void TreeListHeaderWindow::ShowContextMenu(wxContextMenuEvent& event)	// // Menu to choose which columns to display
{
int columns[] = { SHCUT_SHOW_COL_EXT, SHCUT_SHOW_COL_SIZE, SHCUT_SHOW_COL_TIME, SHCUT_SHOW_COL_PERMS, SHCUT_SHOW_COL_OWNER,
								SHCUT_SHOW_COL_GROUP, SHCUT_SHOW_COL_LINK, wxID_SEPARATOR, SHCUT_SHOW_COL_ALL, SHCUT_SHOW_COL_CLEAR };
wxMenu colmenu;
for ( size_t n=0; n < 10; ++n )
	{	MyFrame::mainframe->AccelList->AddToMenu( colmenu, columns[n], wxEmptyString,  (n < 7 ? wxITEM_CHECK : wxITEM_NORMAL) );
	#ifdef __WXX11__
		if ( n < 7 ) colmenu.Check( columns[n], ! IsHidden(n+1) );		// The usual updateui way of doing this doesn't work in X11
	#endif
	}

colmenu.SetTitle( wxT("Columns to Display") );								// This doesn't seem to work with wxGTK-2.4.2, but no harm leaving it in!

wxPoint pt = event.GetPosition();
ScreenToClient(&pt.x, &pt.y);
#ifdef __WXX11__
m_owner->PopupMenu(&colmenu, pt.x, 0);									// In X11 a headerwindow doesn't seem to be able to maintain a popup window, so use the treectrl
#else
PopupMenu(&colmenu, pt.x, pt.y);
#endif
}

void TreeListHeaderWindow::OnColumnSelectMenu(wxCommandEvent& event)
{
bool flag;										// First deal with ShowAll and Clear
if ( ( flag = (event.GetId() == SHCUT_SHOW_COL_ALL) ) ||  (event.GetId() == SHCUT_SHOW_COL_CLEAR) )
	{ for ( size_t col = 1; col < TOTAL_NO_OF_FILEVIEWCOLUMNS; ++col )
				ShowColumn( col, flag );
		return;
	}
														// Otherwise it's a straight-forward column-toggle
size_t col = event.GetId() - SHCUT_SHOW_COL_EXT + 1;								// +1 because Name column is always visible
if ( col > TOTAL_NO_OF_FILEVIEWCOLUMNS ) return;									// Sanity check

ToggleColumn( col );
}

BEGIN_EVENT_TABLE(TreeListHeaderWindow,wxWindow)
    EVT_PAINT         (TreeListHeaderWindow::OnPaint)
    EVT_MOUSE_EVENTS  (TreeListHeaderWindow::OnMouse)
    EVT_SET_FOCUS     (TreeListHeaderWindow::OnSetFocus)
	EVT_MENU_RANGE(SHCUT_SHOW_COL_EXT, SHCUT_SHOW_COL_CLEAR, TreeListHeaderWindow::OnColumnSelectMenu)
	EVT_CONTEXT_MENU(TreeListHeaderWindow::ShowContextMenu)
END_EVENT_TABLE()
//---------------------------------------------------------------------------------------------------------------------------------------------------------------------

MyTreeCtrl::MyTreeCtrl(wxWindow *parentwin, wxWindowID id , const wxPoint& pos, const wxSize& size,
     								          long style, const wxValidator &validator, const wxString& name )
																					    :  wxTreeCtrl(parentwin, id, pos, size, style, validator, name)
{
headerwindow = NULL;																	// To flag to premature OnIdle events that there's nothing safe to do

HeaderimageList = new wxImageList(16, 10, true);						// This holds the down & up symbols for the header of the selected fileview column
HeaderimageList->Add(wxIcon(columnheaderdown_xpm));
HeaderimageList->Add(wxIcon(columnheaderup_xpm));

// >2.6.3 the gtk2 default was changed to no lines, and hard-coded into create! So we need to do this here, not by adjusting style
if ( SHOW_TREE_LINES ) SetWindowStyle( GetWindowStyle() & ~wxTR_NO_LINES );
  else SetWindowStyle( GetWindowStyle() | wxTR_NO_LINES );
	
parent = (MyGenericDirCtrl*)parentwin;

IgnoreRtUp = false;
dragging = false;
startpt.x = startpt.y = 0;

CreateAcceleratorTable();
}

void MyTreeCtrl::CreateAcceleratorTable()
{
wxAcceleratorEntry entries[1];
MyFrame::mainframe->AccelList->FillEntry( entries[0], SHCUT_SELECTALL );	// AcceleratorList supplies the keycode etc
wxAcceleratorTable accel(1, entries);
SetAcceleratorTable(accel);
}

void MyTreeCtrl::RecreateAcceleratorTable()
{
				// It would be nice to be able to empty the old accelerator table before replacing it, but trying caused segfaulting!
CreateAcceleratorTable();																						// Make a new one, with up-to-date data
}
	 
void MyTreeCtrl::OnCtrlA(wxCommandEvent& event)			// Does Select All on Ctrl-A
{
if (parent->fileview==ISLEFT)	 return;		// I can't make it work properly for dir-views, & I'm not sure it's a good idea anyway --- select & delete whole filesystem?
if (!m_current) return;								// Don't ask me why, but if there isn't a current selection, SelectItemRange segfaults


wxTreeItemIdValue cookie; 
wxTreeItemId firstItem = GetFirstChild( GetRootItem(), cookie );					// Use FirstChild as 1st item
if ( ! firstItem.IsOk() ) return;
wxGenericTreeItem* first = (wxGenericTreeItem*)firstItem.m_pItem;

wxTreeItemId lastItem = GetLastChild( GetRootItem() );									// & LastChild as last
if ( ! lastItem.IsOk() ) return;
wxGenericTreeItem* last = (wxGenericTreeItem*)lastItem.m_pItem;

SelectItemRange( first, last );
}

void MyTreeCtrl::OnReceivingFocus(wxFocusEvent& event)		// This tells grandparent window which of the 2 twin-ctrls is the one to consider active
{
if ( ! MyFrame::mainframe->SelectionAvailable ) return;	// If we're constructing/destructing, do nothing lest we break a non-existent window
parent->ParentTab->LeftRight = parent;

wxString selected = parent->GetPath();
parent->DisplaySelectionInTBTextCtrl(selected);				// Take the opportunity also to display current selection in textctrl
MyFrame::mainframe->Layout->OnChangeDir(selected);	//  & the Terminal etc if necessary
if ( parent->fileview == ISLEFT )											// & in the statusbar
	((DirGenericDirCtrl*)parent)->UpdateStatusbarInfo( selected );
 else		((FileGenericDirCtrl*)parent)->UpdateStatusbarInfo( selected );

HelpContext = HCunknown;
event.Skip();
}

void MyTreeCtrl::AdjustForNumlock( wxKeyEvent& event )		// The fix for the X11 Numlock problem
{
event.m_metaDown = FALSE;										// Makes it look as if numlock is off, so Accelerators are properly recognised

if ( event.GetKeyCode() == WXK_ESCAPE )					// This is used in DnD, to abort it if ESC pressed
	MyFrame::mainframe->OnESC_Pressed();

event.Skip();
}

bool MyTreeCtrl::QueryIgnoreRtUp()
{
if ( IgnoreRtUp ) 
		{ IgnoreRtUp = false; return true; }				// If we've been Rt-dragging, and now stopped, ignore the event so that we don't have a spurious context menu
 else return false;														// Otherwise proceed as normal
}

void MyTreeCtrl::OnMiddleDown(wxMouseEvent& event)	// Allows middle-button paste
{
event.Skip();

int flags; wxTreeItemId item = HitTest( event.GetPosition(), flags );	// Find out if we're on a tree item
if ( item.IsOk() ) SelectItem( item );															// If so, select it so that any paste gets deposited there, not in any currently selected folder.  That's _probably_ what's expected

wxCommandEvent cevent;																			// We need a dummy event for parent's method
parent->OnShortcutPaste(cevent);
}

void MyTreeCtrl::OnLtDown(wxMouseEvent& event)	// Start DnD pre-processing if we're near a selection
{
event.Skip();																// First allow base-class to process it

int flags; wxTreeItemId item = HitTest( event.GetPosition(), flags );	// Find out if we're on a tree item
if ( item.IsOk() )	 OnBeginDrag( event );				// If so, if it's a selection start drag-processing
																						// If we're NOT near a selection, ignore
}

void MyTreeCtrl::OnMouseMovement(wxMouseEvent& event)
{
if ( !event.Dragging() )
	{	if ( dragging )	ResetPreDnD();						// If we had been seeing whether or not to drag, & now here's a non-drag event, cancel start-up
		event.Skip(); 	return;										// Either way, skip to give wxTreeCtrl a chance
	}
											// If we're dragging but with the Rt button, do a multiple select if wxTR_MULTIPLE is true
if ( event.RightIsDown() && !event.LeftIsDown() )
	{	if ( ! HasFlag( wxTR_MULTIPLE ) ) return;		// Can't have multiple items selected so abort
		
		int flags; wxArrayTreeItemIds selection;
		wxTreeItemId item = HitTest( event.GetPosition(), flags );	// Find out if we're on the level of a tree item
		bool alreadyaselection = GetSelections(selection);				//  & whether or not the tree already has >0 items selected
		if ( item.IsOk() && ! IsSelected(item) )			// if we're on an item,  & it isn't already selected
			{	if ( ! alreadyaselection ) 							// If there is currently no selection, select it
							SelectItem( item );
					else
#if wxVERSION_NUMBER > 2500
							DoSelectItem( item, false, true );// Otherwise extend the current selection
 #else
				  			SelectItem( item, false, true );
#endif
				IgnoreRtUp = true;
			}
		return; 
	}

if ( dragging ) OnBeginDrag( event );						// Otherwise, if pre-drag has been initiated, go to OnBeginDrag to continue it
}


#if wxVERSION_NUMBER >2503
void MyTreeCtrl::OnRtUp(wxMouseEvent& event)		// In >wxGTK-2.5.3 the wxTreeCtrl has started to interfere with Context Events.  This is a workaround
{
wxContextMenuEvent conevent(wxEVT_CONTEXT_MENU, wxID_ANY, ClientToScreen( event.GetPosition() ) );
if ( parent->fileview == ISLEFT )																				// Send the event to the correct type of parent
	((DirGenericDirCtrl*)parent)->ShowContextMenu( conevent );
 else	
	((FileGenericDirCtrl*)parent)->ShowContextMenu( conevent );
}
#endif

void MyTreeCtrl::OnBeginDrag(wxMouseEvent& event)
{
	// I've taken over the detection of dragging from wxTreeCtrl, which tends to be trigger-happy:  the slightest movement 
	// with a mouse-key down sometimes will trigger it. So I've added a distance requirement.  Note that this is asymmetrical, the x axis requirement being less.
	// This is because of the edge effect:  if the initial click is too near the edge of a pane & the mouse is then moved edgewards, it may go into the adjacent pane
	// before the drag commences, whereupon it picks up the selection in THAT pane, not the original.
	// I've also added a time element
	// There was also a problem with selections:  when a LClick arrives here it hasn't had time to establish a Selection, so we have to assume it will, & check later that is was valid
	
static wxDateTime dragstarttime;	
static wxTreeItemId dragitem;			// We need this static, so that we can check that the item that starts the process is actually selected by the time dragging is triggered
static MyTreeCtrl* originaltree;

if ( startpt.x == 0 && startpt.y == 0 ) 							// 0,0 means not yet set, so this is what we do the first time thru
	{	int flags; dragitem = HitTest( event.GetPosition(), flags );			// Find out if we're near a tree item (& not in the space to the right of it)
		if ( ! dragitem.IsOk()  || flags & wxTREE_HITTEST_ONITEMRIGHT ) 
					return;						// Abort if we're not actually over the selected item.  Otherwise we'll start dragging an old selection even if we're now miles away
		if ( dragitem == GetRootItem() ) return;			// Or if it's root
														// It's a valid LClick, so get things ready for the next time through
		startpt.x = event.GetPosition().x; startpt.y = event.GetPosition().y;
		dragstarttime = wxDateTime::UNow();				// Reset the time to 'now'
		originaltree = this;												// In case we change panes before the drag 'catches'
		dragging = true;	return; 
	}
	
if ( (abs(startpt.x - event.GetPosition().x)) < DnDXTRIGGERDISTANCE 	// If we've not moved far enough in either direction, return
						  &&  (abs(startpt.y - event.GetPosition().y)) < DnDYTRIGGERDISTANCE )			return;

wxTimeSpan diff = wxDateTime::UNow() - dragstarttime;
if ( diff.GetMilliseconds() < 100 )	return;					// If 0.1 sec hasn't passed since dragging was first noticed, return

if ( ! originaltree->IsSelected(dragitem) ) { ResetPreDnD(); return; }	// If we've got this far, we can trigger provided that by now the item has become selected. Otherwise abort the process

													// If we're here, there's a genuine drag event
ResetPreDnD();															// Reset dragging & startpt for the future

if ( MyFrame::mainframe->m_TreeDragMutex.TryLock() != wxMUTEX_NO_ERROR ) 
	{ wxLogTrace( wxT("%s"),wxT("MyTreeCtrl::OnBeginDrag - - - - - - - - - - - - - Mutex was locked")); 
		MyFrame::mainframe->m_TreeDragMutex.Unlock();				// If we don't unlock it here, DnD will never again work.  Surely it will be safe by now . . . .?
		return; 
	}

size_t count;																// No of selections (may be multiple, after all) 
wxArrayString paths;													// Array to store them

count = parent->GetMultiplePaths( paths );				// Fill the array with selected pathname/filenames
if ( ! count ) return;													// No one at home

MyGenericDirCtrl::DnDfilearray = paths;					// Fill the DnD "clipboard" with paths
MyGenericDirCtrl::DnDfilecount = count;					// Store their number
MyGenericDirCtrl::DnDFromID = parent->GetId();		// and the window ID (for refreshing)	Note the reasonable assumption that all selections belong to same pane

MyFrame::mainframe->m_CtrlPressed = event.m_controlDown;	// Store metakey pattern, so that MyFrame::OnBeginDrag will know if we're copying, moving etc 
MyFrame::mainframe->m_ShPressed = event.ShiftDown();
MyFrame::mainframe->m_AltPressed = event.AltDown();

MyFrame::mainframe->OnBeginDrag();						// Pass control to main Drag functions
}

void MyTreeCtrl::OnEndDrag(wxTreeEvent& event)
{
wxYieldIfNeeded();
										// Now we need to find if we've found somewhere to drop onto
wxPoint pt = ScreenToClient( wxGetMousePosition() );						// Locate the mouse?  Screen coords so -> client
int flags;
wxTreeItemId item = HitTest( pt, flags );												// Find out if we're over a valid tree item (& not in the space to the right of it)
if ( item.IsOk()  && !(flags & wxTREE_HITTEST_ONITEMRIGHT) )
	{	wxDirItemData* data = (wxDirItemData*) GetItemData(item);	
		MyGenericDirCtrl::DnDDestfilepath = data->m_path;					// 	so get its path
		MyGenericDirCtrl::DnDToID = parent->GetId();							//	   and its ID (for refreshing)
		if ( parent->fileview == ISRIGHT )												// However, if we're in a fileview pane, need to check that we're not dropping on top of a file
			{	FileData fd( MyGenericDirCtrl::DnDDestfilepath );
				if ( fd.IsSymlinktargetADir( true ) )										// If we're dropping onto a symlink & its target is a dir
					MyGenericDirCtrl::DnDDestfilepath = fd.GetUltimateDestination(); 	// replace path with the symlink target. NB If the target isn't a dir, we don't want to deref the symlink at all; just paste into its parent dir as usual
				 else
					if ( fd.IsValid() && ! fd.IsDir() ) MyGenericDirCtrl::DnDDestfilepath = fd.GetPath();	// If the overall path isn't a dir, remove the filename, so we drop onto the parent dir
			}
	}
 else
	{/*	if ( parent->fileview == ISRIGHT )											// If we're in a fileview pane, but not hovering over something, 
			{	wxString rootpath = parent->startdir;									//  get the root path for the pane
				if ( ! rootpath.IsEmpty() )
					{	MyGenericDirCtrl::DnDDestfilepath = rootpath;				// If it's valid, use it as the path
						MyGenericDirCtrl::DnDToID = parent->GetId();				//	ID (for refreshing)
					}			
			}		**************** I've decided against this, as otherwise all Drops that miss a file will end up in startdir, even if the user thought he was aborting
		else*/
			{	MyGenericDirCtrl::DnDDestfilepath = wxEmptyString;			// If HitTest() missed,  mark invalid 
				MyFrame::mainframe->m_TreeDragMutex.Unlock();	
				return;																				// and bail out
			}
	}
	

enum myDragResult dragtype = MyFrame::mainframe->ParseMetakeys();	// First find out if we've been moving, copying or linking

								// We need to find out if we're Move()ing from an archive: if so, though the code is the same, the clustering is already done
MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane();
bool InArchiveMove = ( active != NULL && active->arcman != NULL && active->arcman->IsArchive() && dragtype==myDragMove );
if ( ! InArchiveMove ) UnRedoManager::StartCluster();

switch ( dragtype )
 {
	case myDragCopy: 			parent->OnPaste( true ); break;
	case myDragMove:  			{DevicesStruct* ds = MyFrame::mainframe->Layout->m_notebook->
																			DeviceMan->QueryAMountpoint( MyGenericDirCtrl::DnDfilearray[0], true );		// See if we're trying to drag from a ro device
												if ( ds != NULL && ds->readonly )  parent->OnPaste( true );	// If we _are_ trying to move from a cdrom or similar, silently alter to Paste
												 else parent->OnDnDMove(); break;}									// Otherwise do a Move as requested
	case myDragHardLink: 
	case myDragSoftLink:		parent->OnLink( true, dragtype ); break;
	case myDragNone:	 		break;
		default:	break;
 }
if ( ! InArchiveMove ) UnRedoManager::EndCluster();
 
MyFrame::mainframe->m_TreeDragMutex.Unlock();								// Finally, release the tree-mutex lock
//wxLogStatus( msg+result );
}

void MyTreeCtrl::OnPaint( wxPaintEvent& event )					// // Copied from wxTreeCtrl only because it's the caller of the overridden (PaintLevel()->) PaintItem
{
if ( parent->fileview == ISLEFT ) return wxTreeCtrl::OnPaint( event );		// // If it's a dirview, use the original version

Col0 = STRIPE_0; Col1 = STRIPE_1;	// // Copy these at the beginning of the paint so, if the defaults get changed halfway thru, we don't end up looking stupid
    wxPaintDC dc(this);
    PrepareDC( dc );

    if ( !m_anchor)
        return;

    dc.SetFont( m_normalFont );
    dc.SetPen( m_dottedPen );

    // this is now done dynamically
    //if(GetImageList() == NULL)
    // m_lineHeight = (int)(dc.GetCharHeight() + 4);

    int y = 2;
    PaintLevel( m_anchor, dc, 0, y, 0 );
}

void MyTreeCtrl::PaintLevel( wxGenericTreeItem *item, wxDC &dc, int level, int &y, int index )	// // I've added index
{
    int x = level*m_indent;
    if (!HasFlag(wxTR_HIDE_ROOT))
    {
        x += m_indent;
    }
    else if (level == 0)
    {
        // always expand hidden root
        int origY = y;
        wxArrayGenericTreeItems& children = item->GetChildren();
        int count = children.Count();
        if (count > 0)
        {
            int n = 0, oldY;
            do {
                oldY = y;
                PaintLevel(children[n], dc, 1, y, n);
            } while (++n < count);

            if (!HasFlag(wxTR_NO_LINES) && HasFlag(wxTR_LINES_AT_ROOT) && count > 0)
            {
                // draw line down to last child
                origY += GetLineHeight(children[0])>>1;
                oldY += GetLineHeight(children[n-1])>>1;
                dc.DrawLine(3, origY, 3, oldY);
            }
        }
        return;
    }

    item->SetX(x+m_spacing);
    item->SetY(y);

    int h = GetLineHeight(item);
    int y_top = y;
    int y_mid = y_top + (h>>1);
    y += h;

    int exposed_x = dc.LogicalToDeviceX(0);
    int exposed_y = dc.LogicalToDeviceY(y_top);

    if (IsExposed(exposed_x, exposed_y, 10000, h))  // 10000 = very much
    {
        const wxPen *pen =
#ifndef __WXMAC__
            // don't draw rect outline if we already have the
            // background color under Mac
            (item->IsSelected() && m_hasFocus) ? wxBLACK_PEN :
#endif // !__WXMAC__
            wxTRANSPARENT_PEN;

        wxColour colText;
        if ( item->IsSelected() )
        {
            colText = wxSystemSettings::GetColour(wxSYS_COLOUR_HIGHLIGHTTEXT);
        }
        else
        {
            wxTreeItemAttr *attr = item->GetAttributes();
            if (attr && attr->HasTextColour())
                colText = attr->GetTextColour();
            else
                colText = GetForegroundColour();
        }

        // prepare to draw
        dc.SetTextForeground(colText);
        dc.SetPen(*pen);

        // draw
        PaintItem(item, dc, index);

        if (HasFlag(wxTR_ROW_LINES))
        {
            // if the background colour is white, choose a
            // contrasting color for the lines
            dc.SetPen(*((GetBackgroundColour() == *wxWHITE)
                         ? wxMEDIUM_GREY_PEN : wxWHITE_PEN));
            dc.DrawLine(0, y_top, 10000, y_top);
            dc.DrawLine(0, y, 10000, y);
        }

        // restore DC objects
        dc.SetBrush(*wxWHITE_BRUSH);
        dc.SetPen(m_dottedPen);
        dc.SetTextForeground(*wxBLACK);

        if (item->HasPlus() && HasButtons())  // should the item show a button?
        {
            if (!HasFlag(wxTR_NO_LINES))
            {
                if (x > (signed)m_indent)
                    dc.DrawLine(x - m_indent, y_mid, x - 5, y_mid);
                else if (HasFlag(wxTR_LINES_AT_ROOT))
                    dc.DrawLine(3, y_mid, x - 5, y_mid);
                dc.DrawLine(x + 5, y_mid, x + m_spacing, y_mid);
            }

            if (m_imageListButtons != NULL)
            {
                // draw the image button here
                int image_h = 0, image_w = 0, image = wxTreeItemIcon_Normal;
                if (item->IsExpanded()) image = wxTreeItemIcon_Expanded;
                if (item->IsSelected())
                    image += wxTreeItemIcon_Selected - wxTreeItemIcon_Normal;
                m_imageListButtons->GetSize(image, image_w, image_h);
                int xx = x - (image_w>>1);
                int yy = y_mid - (image_h>>1);
                dc.SetClippingRegion(xx, yy, image_w, image_h);
                m_imageListButtons->Draw(image, dc, xx, yy,
                                         wxIMAGELIST_DRAW_TRANSPARENT);
                dc.DestroyClippingRegion();
            }

#if wxVERSION_NUMBER < 2501
            else if (HasFlag(wxTR_TWIST_BUTTONS))
            {
                // draw the twisty button here

                if (HasFlag(wxTR_AQUA_BUTTONS))
                {
                    // This causes update problems, so disabling for now.
#if 0 //#ifdef __WXMAC__
                    wxMacPortSetter helper(&dc) ;
                    wxMacWindowClipper clipper(this) ;
                    wxDC::MacSetupBackgroundForCurrentPort( MacGetBackgroundBrush() ) ;

                    int loc_x = x - 5 ;
                    int loc_y = y_mid - 6 ;
                    MacWindowToRootWindow( & loc_x , & loc_y ) ;
                    Rect bounds = { loc_y , loc_x , loc_y + 18 , loc_x + 12 } ;
                    ThemeButtonDrawInfo info = { kThemeStateActive , item->IsExpanded() ? kThemeDisclosureDown : kThemeDisclosureRight ,
                        kThemeAdornmentNone };
                    DrawThemeButton( &bounds, kThemeDisclosureButton ,
                        &info , NULL , NULL , NULL , NULL ) ;
#else
	#if wxVERSION_NUMBER < 2500	// // Kludge to get 2.5.0 to compile. It doesn't seem to matter anyway!?
                    if (item->IsExpanded())
                        dc.DrawBitmap( *m_arrowDown, x-5, y_mid-6, true );
                    else
                        dc.DrawBitmap( *m_arrowRight, x-5, y_mid-6, true );
	#endif
#endif
                }
                else
                {
                    dc.SetBrush(*m_hilightBrush);
                    dc.SetPen(*wxBLACK_PEN);
                    wxPoint button[3];

                    if (item->IsExpanded())
                    {
                        button[0].x = x-5;
                        button[0].y = y_mid-2;
                        button[1].x = x+5;
                        button[1].y = y_mid-2;
                        button[2].x = x;
                        button[2].y = y_mid+3;
                    }
                    else
                    {
                        button[0].y = y_mid-5;
                        button[0].x = x-2;
                        button[1].y = y_mid+5;
                        button[1].x = x-2;
                        button[2].y = y_mid;
                        button[2].x = x+3;
                    }
                    dc.DrawPolygon(3, button);
                    dc.SetPen(m_dottedPen);
                }
            }
            else // if (HasFlag(wxTR_HAS_BUTTONS))
            {
                // draw the plus sign here
                dc.SetPen(*wxGREY_PEN);
                dc.SetBrush(*wxWHITE_BRUSH);
                dc.DrawRectangle(x-5, y_mid-4, 11, 9);
                dc.SetPen(*wxBLACK_PEN);
                dc.DrawLine(x-2, y_mid, x+3, y_mid);
                if (!item->IsExpanded())
                    dc.DrawLine(x, y_mid-2, x, y_mid+3);
                dc.SetPen(m_dottedPen);
            }
        }
        else if (!HasFlag(wxTR_NO_LINES))  // no button; maybe a line?
        {
            // draw the horizontal line here
            int x_start = x;
            if (x > (signed)m_indent)
                x_start -= m_indent;
            else if (HasFlag(wxTR_LINES_AT_ROOT))
                x_start = 3;
            dc.DrawLine(x_start, y_mid, x + m_spacing, y_mid);
        }
    } //else wxLogTrace("Hidden  %u   y = %d   yexposed = %d    %s", index, item->GetY(), exposed_y, item->GetText().c_str() );

#else													// If >=2.5.1
		}
	}
           else // no custom buttons
            {
                static const int wImage = 9;
                static const int hImage = 9;
                
                int flag = 0;
                if (item->IsExpanded())
                    flag |= wxCONTROL_EXPANDED;
                if (item == m_underMouse)
                    flag |= wxCONTROL_CURRENT;
                                            
                wxRendererNative::Get().DrawTreeItemButton
                                        (
                                            this,
                                            dc,
                                            wxRect(x - wImage/2,
                                                   y_mid - hImage/2,
                                                   wImage, hImage),
                                            flag
                                        );
            }
#endif		//  >= 2.5.1


    if (item->IsExpanded())
    {
        wxArrayGenericTreeItems& children = item->GetChildren();
        int count = children.Count();
        if (count > 0)
        {
            int n = 0, oldY;
            ++level;
            do {
                oldY = y;
                PaintLevel(children[n], dc, level, y);
            } while (++n < count);

            if (!HasFlag(wxTR_NO_LINES) && count > 0)
            {
                // draw line down to last child
                oldY += GetLineHeight(children[n-1])>>1;
                if (HasButtons()) y_mid += 5;

                // Only draw the portion of the line that is visible, in case it is huge
                wxCoord	xOrigin=0, yOrigin=0, width, height;
                dc.GetDeviceOrigin(&xOrigin, &yOrigin);
                yOrigin = abs(yOrigin);
                GetClientSize(&width, &height);

                // Move end points to the begining/end of the view?
                if (y_mid < yOrigin)
                    y_mid = yOrigin;
                if (oldY > yOrigin + height)
                    oldY = yOrigin + height;

                // after the adjustments if y_mid is larger than oldY then the line
                // isn't visible at all so don't draw anything
                if (y_mid < oldY)
                    dc.DrawLine(x, y_mid, x, oldY);
            }
        }
    }
}

void MyTreeCtrl::PaintItem(wxGenericTreeItem *item, wxDC& dc, int index)
{
static const int NO_IMAGE = -1;																	// //
    // TODO implement "state" icon on items

    wxTreeItemAttr *attr = item->GetAttributes();
    if ( attr && attr->HasFont() )
        dc.SetFont(attr->GetFont());
    else if (item->IsBold())
        dc.SetFont(m_boldFont);

    int text_w = 0, text_h = 0;
    
    dc.GetTextExtent( item->GetText(), &text_w, &text_h );

    int total_h = GetLineHeight(item);

    if ( item->IsSelected() )
    {
        dc.SetBrush(*(m_hasFocus ? m_hilightBrush : m_hilightUnfocusedBrush));
    }
    else
    {
        wxColour colBg;
        if ( attr && attr->HasBackgroundColour() )
            colBg = attr->GetBackgroundColour();
        else
		  if ( USE_STRIPES )																					// // If we want the fileview background to be striped with alternate colours
              colBg = index % 2 ? Col1 : Col0;
		   else colBg = m_backgroundColour;

        dc.SetBrush(wxBrush(colBg, wxSOLID));
    }

    int offset = HasFlag(wxTR_ROW_LINES) ? 1 : 0;

    dc.DrawRectangle(0, item->GetY()+offset,
		     headerwindow->GetWidth(),																	// //
		     total_h-offset);

    dc.SetBackgroundMode(wxTRANSPARENT);
    int extraH = (total_h > text_h) ? (total_h - text_h)/2 : 0;
    int extra_offset = 0;

DataBase* stat = NULL;																						// //
					// // After simplifying (& so probably speeding up) DirGenericDirCtrl::ReloadTree(), I started getting rare Asserts where index==GetCount() during Moving/Pasting
					// // I suspect data is sometimes being added to the treectrl faster than to the array, If a refresh then occurs at the wrong time . . .
if ( ((FileGenericDirCtrl*)parent)->CombinedFileDataArray.GetCount() <= (size_t)index )  return;	// //
stat = &((FileGenericDirCtrl*)parent)->CombinedFileDataArray.Item(index);			// // Get a ptr to the current stat data
		
for(size_t i = 0; i < headerwindow->GetColumnCount(); ++i) 								// //
 {
if ( headerwindow->IsHidden(i) ) continue;															// // Not a lot to do for a hidden column

	int coord_x = extra_offset, image_x = coord_x;
	int clip_width = headerwindow->GetColumnWidth(i);										// //
	int image_h = 0, image_w = 0; //2;
	int image = NO_IMAGE;
	
	if(i == 0) {																										// //
	    image = item->GetCurrentImage();
	    coord_x = item->GetX();
					}
	else {
	    image = NO_IMAGE;	// //item->GetImage(i);
			}
	
	if(image != NO_IMAGE) 
	 {
	    if(m_imageListNormal) {
		m_imageListNormal->GetSize( image, image_w, image_h );
		image_w += 4;
	    										}
	    else {
		image = NO_IMAGE;
	    		}
	 }

	// honor text alignment
	wxString text( wxT("") );
	if ( !i)																											// // If col 0, do it the standard way
		 text = item->GetText();
     else 																											// // The whole 'else' is mine
		{	wxDateTime time;
			if ( stat != NULL && ! headerwindow->IsHidden(i) )
				switch( i )
				{	case ext:					text = (stat->GetFilename().Mid(1)).AfterFirst(wxT('.')); break;
				
					case filesize:				text = stat->GetParsedSize();
														break;
					case modtime:			time.Set( stat->ModificationTime() ); text = time.Format(wxT("%d/%m/%y %H:%M"));
														break;
					case permissions:		text = stat->PermissionsToText();
														break;
					case owner:				text = stat->GetOwner(); if ( text.IsEmpty() )	text.Printf(wxT("%u"),  stat->OwnerID());
														break;
					case group:				text = stat->GetGroup(); if ( text.IsEmpty() )	text.Printf(wxT("%u"),  stat->GroupID());
														break;
					case linkage:				if ( stat->IsSymlink() )
															{	if ( stat->GetSymlinkData() && stat->GetSymlinkData()->IsValid() )
																	text.Printf( wxT("-->%s"), stat->GetSymlinkDestination( true ).c_str() );
																 else text = wxT(" *** Broken symlink ***");
															}
				}		
		}
		
	switch(headerwindow->GetColumn(i).GetAlignment()) {
	case wxTL_ALIGN_LEFT:
	    coord_x += image_w + 2;
	    image_x = coord_x - image_w;
	    break;
	case wxTL_ALIGN_RIGHT:
	    dc.GetTextExtent(text, &text_w, NULL);
	    coord_x += clip_width - text_w - image_w - 2;
	    image_x = coord_x - image_w;
	    break;
	case wxTL_ALIGN_CENTER:
	    dc.GetTextExtent(text, &text_w, NULL);
	    //coord_x += (clip_width - text_w)/2 + image_w;
	    image_x += (clip_width - text_w - image_w)/2 + 2;
	    coord_x = image_x + image_w;
	}	

	wxDCClipper clipper(dc, /*coord_x,*/ extra_offset,
			    item->GetY() + extraH, clip_width,
			    total_h);

	if(image != NO_IMAGE) {
	    m_imageListNormal->Draw( image, dc, image_x,
				     item->GetY() +((total_h > image_h)?
						    ((total_h-image_h)/2):0),
				     wxIMAGELIST_DRAW_TRANSPARENT );
	}

	dc.DrawText( text,
		     (wxCoord)(coord_x /*image_w + item->GetX()*/),
		     (wxCoord)(item->GetY() + extraH));
	extra_offset += headerwindow->GetColumnWidth(i);
 }

    // restore normal font
    dc.SetFont( m_normalFont );
}


void MyTreeCtrl::OnIdle( wxIdleEvent& event )	// // Makes sure any change in column width is reflected within
{
if (headerwindow==NULL) { event.Skip(); return; }				// Necessary as event may be called before there IS a headerwindow	

if (headerwindow->m_dirty) 													// Although MyTreeCtrl has an m_dirty too, that's just there because TreeListHeaderWindow writes to it!
	{	headerwindow->m_dirty = false;									// The one that is changed when columns are resized is the TreeListHeaderWindow one
		Refresh();
	}

event.Skip();																			// Without this, gtk1 has focus problems; and in X11 the scrollbars get stuck in the top-left corner on creation!
}

void MyTreeCtrl::OnSize( wxSizeEvent& event )	// // Keeps the treectrl columns in step with the headerwindows
{
#if wxVERSION_NUMBER > 2500
if (headerwindow != NULL) headerwindow->m_dirty = true;	// In >=2.5.1, the headerwindows don't otherwise seem to get refreshed on splitter dragging
#endif

AdjustMyScrollbars();

event.Skip();
}

void MyTreeCtrl::AdjustMyScrollbars()	// // Taken from src/generic/treectlg.cpp, because of the headerwindow line
{
const int PIXELS_PER_UNIT = 10;	// // This was set in src/generic/treectlg.cpp

if (m_anchor)
    {
        int x = 0, y = 0;
        m_anchor->GetSize( x, y, this );
		if (headerwindow != NULL)  x = headerwindow->GetTotalPreferredWidth();	// // This means we get the column width, not just that of m_anchor's name

		y += PIXELS_PER_UNIT+2; // one more scrollbar unit + 2 pixels
      // //  x += PIXELS_PER_UNIT+2; // one more scrollbar unit + 2 pixels                This actually makes things worse
        int x_pos = GetScrollPos( wxHORIZONTAL );
        int y_pos = GetScrollPos( wxVERTICAL );
        SetScrollbars( PIXELS_PER_UNIT, PIXELS_PER_UNIT, x/PIXELS_PER_UNIT, y/PIXELS_PER_UNIT, x_pos, y_pos );
    }
 else
    {
        SetScrollbars( 0, 0, 0, 0 );
    }
}

void MyTreeCtrl::OnScroll( wxScrollWinEvent& event )	// // Keeps the headerwindows in step with the treectrl colums
{
if (headerwindow != NULL) headerwindow->Refresh();
event.Skip();
}

void MyTreeCtrl::OnEraseBackground( wxEraseEvent& event )	// This is only needed in gtk2 under kde and brain-dead theme-manager, but it doesn't seem to do any harm elsewhere
{
wxClientDC* clientDC = NULL;  // We may or may not get a valid dc from the event, so be prepared
if ( !event.GetDC() ) clientDC = new wxClientDC(this);
wxDC* dc = clientDC ? clientDC : event.GetDC();

wxColour colBg = GetBackgroundColour();  // Use the chosen background if there is one
if ( ! colBg.Ok() ) colBg = wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW);

dc->SetBackground(wxBrush(colBg, wxSOLID));
dc->Clear();

if ( clientDC ) delete clientDC;
}

BEGIN_EVENT_TABLE(MyTreeCtrl,wxTreeCtrl)
		EVT_KEY_DOWN( MyTreeCtrl::AdjustForNumlock )
		EVT_SET_FOCUS(MyTreeCtrl::OnReceivingFocus)	
		EVT_MENU(SHCUT_SELECTALL,MyTreeCtrl::OnCtrlA)
#if wxVERSION_NUMBER >2503
		EVT_RIGHT_DOWN(MyTreeCtrl::OnRtDown)					// Just swallows the event, to stop it triggering an unwanted context menu in >GTK-2.5.3
		EVT_RIGHT_UP(MyTreeCtrl::OnRtUp)								// Takes over the context menu generation
#endif
		EVT_LEFT_DOWN(MyTreeCtrl::OnLtDown)
		EVT_MOTION(MyTreeCtrl::OnMouseMovement)
		EVT_MIDDLE_DOWN(MyTreeCtrl::OnMiddleDown)			// Allows middle-button paste
		EVT_LEAVE_WINDOW(MyTreeCtrl::OnLeavingWindow)
		EVT_PAINT          (MyTreeCtrl::OnPaint)
		EVT_ERASE_BACKGROUND(MyTreeCtrl::OnEraseBackground)
		EVT_IDLE           (MyTreeCtrl::OnIdle)
		EVT_SIZE			(MyTreeCtrl::OnSize)
		EVT_SCROLLWIN(MyTreeCtrl::OnScroll)
END_EVENT_TABLE()

//---------------------------------------------------------------------------------------------------------------------------------------------------------------------




